# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools

import numpy as np

from ...config import options
from ...serialization.serializables import FieldTypes, StringField, TupleField
from ..core import TensorOrder
from ..operands import TensorOperand, TensorOperandMixin
from ..utils import decide_chunk_sizes, normalize_shape


class TensorDataSource(TensorOperand, TensorOperandMixin):
    """
    Tensor data source base class, provide universal tile logic,
    subclass can overwrite tile method.
    """

    __slots__ = ()

    def to_chunk_op(self, *args):
        chunk_shape = args[0]
        chunk_op = self.copy().reset_key()
        chunk_op.extra_params = {"size": chunk_shape}  # to make op key different
        return chunk_op

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        chunk_size = tensor.extra_params.raw_chunk_size or options.chunk_size
        chunk_size = decide_chunk_sizes(tensor.shape, chunk_size, tensor.dtype.itemsize)
        chunk_size_idxes = (range(len(size)) for size in chunk_size)

        out_chunks = []
        for chunk_shape, chunk_idx in zip(
            itertools.product(*chunk_size), itertools.product(*chunk_size_idxes)
        ):
            chunk_op = op.to_chunk_op(chunk_shape, chunk_idx, chunk_size)
            out_chunk = chunk_op.new_chunk(
                None, shape=chunk_shape, index=chunk_idx, order=tensor.order
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            tensor.shape,
            chunks=out_chunks,
            nsplits=chunk_size,
            order=tensor.order,
            **tensor.extra_params
        )


class TensorNoInput(TensorDataSource):
    """
    Tensor operand with no inputs.
    """

    def check_inputs(self, inputs):
        # no inputs
        if inputs and len(inputs) > 0:
            raise ValueError("Tensor data source has no inputs")

    def _new_chunks(self, inputs, kws=None, **kw):
        shape = kw.get("shape", None)
        self.extra_params[
            "shape"
        ] = shape  # set shape to make the operand key different
        return super()._new_chunks(inputs, kws=kws, **kw)

    def _new_tileables(self, inputs, kws=None, **kw):
        shape = kw.get("shape", None)
        self.extra_params[
            "shape"
        ] = shape  # set shape to make the operand key different
        return super()._new_tileables(inputs, kws=kws, **kw)

    def __call__(self, shape, chunk_size=None, order=None):
        shape = normalize_shape(shape)
        order = TensorOrder.C_ORDER if order is None else order
        return self.new_tensor(None, shape, raw_chunk_size=chunk_size, order=order)


class TensorHasInput(TensorDataSource):
    """
    Tensor operand with a single input.
    """

    @property
    def input(self):
        return self._input

    def check_inputs(self, inputs):
        # no inputs
        if len(inputs) != 1:
            raise ValueError("Tensor can only have 1 input")

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        output = op.outputs[0]

        out_chunks = []
        for c in op.input.chunks:
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk([c], shape=c.shape, index=c.index, order=output.order)
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            output.shape,
            order=output.order,
            chunks=out_chunks,
            nsplits=op.input.nsplits,
        )

    def __call__(self, a, order=None):
        order = a.order if order is None else order
        return self.new_tensor([a], a.shape, order=order)


class TensorLike(TensorHasInput):
    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.dtype is None:
            self.dtype = self.input.dtype
        if self.gpu is None:
            self.gpu = self.input.op.gpu

        # FIXME: remove when cupy supports other dtypes
        if self.gpu and self.dtype not in (np.float32, np.float64):
            raise NotImplementedError(
                "Sparse tensor on GPU only supports float32 and float64"
            )


class TensorFromHDF5Like(TensorNoInput):
    _filename = StringField("filename")
    _group = StringField("group")
    _dataset = StringField("dataset")
    _axis_offsets = TupleField("axis_offsets", FieldTypes.int64)

    def __init__(self, filename=None, group=None, dataset=None, **kw):
        super().__init__(_filename=filename, _group=group, _dataset=dataset, **kw)

    @property
    def filename(self):
        return self._filename

    @property
    def group(self):
        return self._group

    @property
    def dataset(self):
        return self._dataset

    @property
    def axis_offsets(self):
        return self._axis_offsets

    @property
    def path(self):
        return self.get_path(self.group, self.dataset)

    def to_chunk_op(self, *args):
        _, chunk_index, nsplits = args
        chunk_op = super().to_chunk_op(*args)
        cum_offsets = [[0] + np.cumsum(ns).tolist() for ns in nsplits]
        axis_offsets = []
        for axis, idx in enumerate(chunk_index):
            axis_offsets.append(cum_offsets[axis][idx])
        chunk_op._axis_offsets = tuple(axis_offsets)
        return chunk_op

    @staticmethod
    def get_path(group, dataset):
        paths = []
        if group:
            paths.append(group)
        paths.append(dataset)
        return "/".join(paths)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import pickle
import uuid
from io import BytesIO
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from xoscar.backends.communication.utils import _convert_to_cupy_ndarray
from xoscar.serialization import deserialize, serialize

from ..utils import implements, lazy_import
from .base import ObjectInfo, StorageBackend, StorageLevel, register_storage_backend
from .core import StorageFileObject

cupy = lazy_import("cupy")
cudf = lazy_import("cudf")


_id_to_buffers = dict()


class CudaFileObject:
    def __init__(self, mode: str, object_id: str, size: int = None):
        self._mode = mode
        self._object_id = object_id
        self._size = size
        self._closed = False
        self._buffers = None
        self._headers = None
        self._offset = None
        # for read
        self._has_read_headers = None
        # for write
        self._has_write_headers = None
        self._cur_buffer_index = None
        if "r" in mode:
            assert object_id is not None
            self._initialize_read()
        elif "w" in mode:
            self._initialize_write()

    @property
    def object_id(self):
        return self._object_id

    @property
    def mode(self):
        return self._mode

    def set_buffers_by_sizes(self, sizes: List[int]):
        from rmm import DeviceBuffer

        self._buffers = [
            _convert_to_cupy_ndarray(DeviceBuffer(size=size)) for size in sizes
        ]

    @property
    def header(self):
        return self._headers

    @header.setter
    def header(self, header):
        self._headers = header

    @property
    def buffer(self):
        return self._buffers

    def _initialize_read(self):
        from cudf.core.buffer import Buffer

        self._offset = 0
        self._has_read_headers = False
        self._buffers = []
        (metas, serialized), buffers = _id_to_buffers[self._object_id]
        self._headers = headers = (metas.copy(), serialized)
        buffer_types = []
        for buf in buffers:
            if isinstance(buf, cupy.ndarray):
                self._buffers.append(buf.astype("u1", copy=False))
                buffer_types.append(["cuda", buf.size])
            elif isinstance(buf, Buffer):
                self._buffers.append(_convert_to_cupy_ndarray(buf))
                buffer_types.append(["cuda", buf.size])
            else:
                size = getattr(buf, "size", len(buf))
                self._buffers.append(buf)
                buffer_types.append(["memory", size])
        headers[0]["buffer_types"] = buffer_types

    def _initialize_write(self):
        self._had_write_headers = False
        self._cur_buffer_index = 0
        self._buffers = []
        self._offset = 0

    def read(self, size: int):
        # we read cuda_header first and then read cuda buffers one by one,
        # the return value's size is not exactly the specified size.
        from cudf.core.buffer import Buffer
        from cupy.cuda import MemoryPointer
        from cupy.cuda.memory import UnownedMemory

        if not self._has_read_headers:
            self._has_read_headers = True
            return pickle.dumps(self._headers)
        if len(self._buffers) == 0:
            return ""
        cur_buf = (
            self._buffers[0].getbuffer()
            if isinstance(self._buffers[0], BytesIO)
            else self._buffers[0]
        )
        cur_size = getattr(cur_buf, "size", len(cur_buf))

        if size >= cur_size - self._offset:
            if isinstance(cur_buf, UnownedMemory):
                cupy_pointer = MemoryPointer(cur_buf, self._offset)
                content = Buffer(cupy_pointer.ptr, size=cur_buf.size - self._offset)
            else:
                content = cur_buf[self._offset : self._offset + size]
            self._offset = 0
            self._buffers.pop(0)
            return content
        else:
            if isinstance(cur_buf, UnownedMemory):
                cupy_pointer = MemoryPointer(cur_buf, self._offset)
                self._offset += size
                return Buffer(cupy_pointer.ptr, size=size)
            else:
                content = cur_buf[self._offset : self._offset + size]
                self._offset += size
                return content

    def write(self, content):
        from cupy.cuda import MemoryPointer
        from cupy.cuda.memory import UnownedMemory
        from rmm import DeviceBuffer

        if not self._has_write_headers:
            self._headers = headers = pickle.loads(content)
            buffer_types = headers[0]["buffer_types"]
            for buffer_type, size in buffer_types:
                if buffer_type == "cuda":
                    self._buffers.append(DeviceBuffer(size=size))
                else:
                    self._buffers.append(BytesIO(b"0" * size))
            self._has_write_headers = True
            return

        cur_buf = self._buffers[self._cur_buffer_index]
        cur_buf_size = self._headers[0]["buffer_types"][self._cur_buffer_index][1]
        if isinstance(cur_buf, DeviceBuffer):
            cur_cupy_memory = UnownedMemory(cur_buf.ptr, cur_buf.size, cur_buf)
            cupy_pointer = MemoryPointer(cur_cupy_memory, self._offset)

            if isinstance(content, bytes):
                content_length = len(content)
                source_mem = np.frombuffer(content, dtype="uint8").ctypes.data_as(
                    ctypes.c_void_p
                )
            else:
                source_mem = content.data
                content_length = content.size
            cupy_pointer.copy_from(source_mem, content_length)
        else:
            content_length = len(content)
            cur_buf.write(content)
        if content_length + self._offset >= cur_buf_size:
            if isinstance(cur_buf, BytesIO):
                self._buffers[self._cur_buffer_index] = cur_buf.getvalue()
            self._cur_buffer_index += 1
            self._offset = 0
        else:
            self._offset += content_length

    def _read_close(self):
        self._offset = None
        self._cuda_buffers = None
        self._cuda_header = None
        self._has_read_headers = None

    def _write_close(self):
        headers = self._headers
        headers[0].pop("buffer_types")
        # hold cuda buffers

        _id_to_buffers[self._object_id] = headers, self._buffers

        self._has_write_headers = None
        self._cur_buffer_index = None
        self._cuda_buffers = None
        self._cuda_header = None
        self._offset = None

    def close(self):
        if self._closed:
            return
        self._closed = True
        if self._mode == "w":
            self._write_close()
        else:
            self._read_close()


@register_storage_backend
class CudaStorage(StorageBackend):
    name = "cuda"
    is_seekable = False

    def __init__(self, size=None):
        self._size = size

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        size = kwargs.pop("size", None)
        if kwargs:  # pragma: no cover
            raise TypeError(f'CudaStorage got unexpected config: {",".join(kwargs)}')

        return dict(size=size), dict()

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        pass

    @property
    @implements(StorageBackend.level)
    def level(self):
        return StorageLevel.GPU

    @property
    @implements(StorageBackend.size)
    def size(self) -> Union[int, None]:
        return self._size

    @staticmethod
    def _to_cuda(obj):  # pragma: no cover
        if isinstance(obj, np.ndarray):
            return cupy.asarray(obj)
        elif isinstance(obj, pd.DataFrame):
            return cudf.DataFrame.from_pandas(obj)
        elif isinstance(obj, pd.Series):
            return cudf.Series.from_pandas(obj)
        return obj

    @implements(StorageBackend.get)
    async def get(self, object_id: str, **kwargs) -> object:
        if kwargs:
            raise NotImplementedError(f'Got unsupported args: {",".join(kwargs)}')

        from cudf.core.buffer import Buffer as CPBuffer
        from rmm import DeviceBuffer

        headers, buffers = _id_to_buffers[object_id]
        new_buffers = []
        for buf in buffers:
            if isinstance(buf, cupy.ndarray):
                new_buffers.append(DeviceBuffer(ptr=buf.data.ptr, size=buf.size))
            elif isinstance(buf, CPBuffer):
                new_buffers.append(DeviceBuffer(ptr=buf.ptr, size=buf.size))
            else:
                new_buffers.append(buf)
        return deserialize(headers, new_buffers)

    @implements(StorageBackend.put)
    async def put(self, obj, importance=0) -> ObjectInfo:
        from cudf.core.buffer import Buffer as CPBuffer

        string_id = str(uuid.uuid4())
        headers, buffers = serialize(obj)
        size = sum(
            buf.size for buf in buffers if isinstance(buf, (cupy.ndarray, CPBuffer))
        )
        _id_to_buffers[string_id] = headers, buffers
        return ObjectInfo(size=size, object_id=string_id)

    @implements(StorageBackend.delete)
    async def delete(self, object_id: str):
        if object_id in _id_to_buffers:
            del _id_to_buffers[object_id]

    @implements(StorageBackend.object_info)
    async def object_info(self, object_id: str) -> ObjectInfo:
        from cudf.core.buffer import Buffer as CPBuffer

        size = sum(
            buf.size
            for buf in _id_to_buffers[object_id][1]
            if isinstance(buf, (cupy.ndarray, CPBuffer))
        )
        return ObjectInfo(size=size, object_id=object_id)

    @implements(StorageBackend.open_writer)
    async def open_writer(self, size=None) -> StorageFileObject:
        object_id = str(uuid.uuid4())
        cuda_writer = CudaFileObject(object_id=object_id, mode="w", size=size)
        return StorageFileObject(cuda_writer, object_id=object_id)

    @implements(StorageBackend.open_reader)
    async def open_reader(self, object_id) -> StorageFileObject:
        cuda_reader = CudaFileObject(mode="r", object_id=object_id)
        return StorageFileObject(cuda_reader, object_id=object_id)

    @implements(StorageBackend.list)
    async def list(self) -> List:  # pragma: no cover
        raise NotImplementedError("Cuda storage doesn't support `list` method.")



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Set

from .....core import TileableData
from .....dataframe.core import BaseDataFrameData, BaseSeriesData
from .....dataframe.groupby.aggregation import DataFrameGroupByAgg
from .....dataframe.indexing.getitem import DataFrameIndex
from .....dataframe.indexing.setitem import DataFrameSetitem
from .....dataframe.merge import DataFrameMerge
from .....typing import OperandType
from .utils import get_cols_exclude_index


class SelfColumnSelector:
    _OP_TO_SELECT_FUNCTION = {}

    @classmethod
    def register(
        cls,
        op_cls: OperandType,
        func: Callable[[TileableData], Set[Any]],
        replace: bool = False,
    ) -> None:
        if op_cls not in cls._OP_TO_SELECT_FUNCTION or replace:
            cls._OP_TO_SELECT_FUNCTION[op_cls] = func
        else:
            raise ValueError(f"key {op_cls} exists.")

    @classmethod
    def select(cls, tileable_data: TileableData) -> Set[Any]:
        """
        TODO: docstring
        """
        op_type = type(tileable_data.op)
        if op_type in cls._OP_TO_SELECT_FUNCTION:
            return cls._OP_TO_SELECT_FUNCTION[op_type](tileable_data)
        for op_cls in op_type.__mro__:
            if op_cls in cls._OP_TO_SELECT_FUNCTION:
                cls._OP_TO_SELECT_FUNCTION[op_type] = cls._OP_TO_SELECT_FUNCTION[op_cls]
                return cls._OP_TO_SELECT_FUNCTION[op_cls](tileable_data)
        return set()


def register_selector(op_type: OperandType) -> Callable:
    def wrap(selector_func: Callable):
        SelfColumnSelector.register(op_type, selector_func)
        return selector_func

    return wrap


@register_selector(DataFrameSetitem)
def df_setitem_select_function(tileable_data: TileableData) -> Set[Any]:
    if isinstance(tileable_data.op.indexes, list):
        return set(tileable_data.op.indexes)
    else:
        return {tileable_data.op.indexes}


@register_selector(DataFrameIndex)
def df_getitem_select_function(tileable_data: TileableData) -> Set[Any]:
    if tileable_data.op.col_names is not None:
        col_names = tileable_data.op.col_names
        if isinstance(col_names, list):
            return set(tileable_data.op.col_names)
        else:
            return {tileable_data.op.col_names}
    else:
        if isinstance(tileable_data, BaseDataFrameData):
            return set(tileable_data.dtypes.index)
        elif isinstance(tileable_data, BaseSeriesData):
            return {tileable_data.name}


@register_selector(DataFrameGroupByAgg)
def df_groupby_agg_select_function(tileable_data: TileableData) -> Set[Any]:
    """
    Make sure the "group by columns" are preserved.
    """

    op: DataFrameGroupByAgg = tileable_data.op
    by = op.groupby_params["by"]

    if isinstance(tileable_data, BaseDataFrameData):
        return get_cols_exclude_index(tileable_data, by)
    elif isinstance(tileable_data, BaseSeriesData):
        return {tileable_data.name}
    else:
        return set()


@register_selector(DataFrameMerge)
def df_merge_select_function(tileable_data: TileableData) -> Set[Any]:
    """
    Make sure the merge keys are preserved.
    """

    op: DataFrameMerge = tileable_data.op
    on = op.on
    if on is not None:
        return get_cols_exclude_index(tileable_data, on)

    ret = set()
    left_data: BaseDataFrameData = op.inputs[0]
    right_data: BaseDataFrameData = op.inputs[1]
    left_index = op.left_index
    right_index = op.right_index
    left_on = op.left_on if isinstance(op.left_on, list) else [op.left_on]
    right_on = op.right_on if isinstance(op.right_on, list) else [op.right_on]

    if left_index and right_index:
        return ret

    if left_index:
        for col in right_data.dtypes.index:
            if col in right_on:
                ret.add(col)
        return ret
    if right_index:
        for col in left_data.dtypes.index:
            if col in left_on:
                ret.add(col)
        return ret

    for data, merge_keys, suffix in zip(
        [left_data, right_data], [left_on, right_on], op.suffixes
    ):
        if merge_keys is None:
            continue
        for col in data.dtypes.index:
            if col in merge_keys:
                other_data = right_data if data is left_data else left_data
                other_merge_keys = right_on if merge_keys is left_on else left_on

                if col in other_data.dtypes.index and col not in other_merge_keys:
                    # if the merge key exists in the other dataframe but not in the other
                    # dataframe's merge keys, suffixes will be added.
                    # TODO: this does not work when col is a tuple.
                    suffix_col = str(col) + suffix
                    ret.add(suffix_col)
                else:
                    ret.add(col)
    return ret



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob as glob_
import os
from typing import Dict, List
from urllib.parse import urlparse

from ..compression import compress
from .base import FileSystem, path_type
from .local import LocalFileSystem
from .oss import OSSFileSystem

_filesystems = {"file": LocalFileSystem, "oss": OSSFileSystem}
_scheme_to_dependencies = {
    "hdfs": ["pyarrow"],
    "az": ["fsspec", "adlfs"],
    "abfs": ["fsspec", "adlfs"],
    "s3": ["fsspec", "s3fs"],
}


def register_filesystem(name: str, fs):
    _filesystems[name] = fs


def get_scheme(path: path_type) -> str:
    if os.path.exists(path) or glob_.glob(path):
        scheme = "file"
    else:
        scheme = urlparse(path).scheme
    if scheme == "" or len(scheme) == 1:  # len == 1 for windows
        scheme = "file"
    return scheme


def get_fs(path: path_type, storage_options: Dict = None) -> FileSystem:
    if storage_options is None:
        storage_options = dict()

    scheme = get_scheme(path)
    if scheme in _filesystems:
        file_system_type = _filesystems[scheme]
        if scheme == "file" or scheme == "oss":
            # local file systems are singletons.
            return file_system_type.get_instance()
        else:
            storage_options = file_system_type.get_storage_options(
                storage_options, path
            )
            return file_system_type(**storage_options)
    elif scheme in _scheme_to_dependencies:  # pragma: no cover
        dependencies = ", ".join(_scheme_to_dependencies[scheme])
        raise ImportError(f"Need to install {dependencies} to access {scheme}.")
    else:
        raise ValueError(
            f"Unknown file system type: {scheme}, "
            f'available include: {", ".join(_scheme_to_dependencies.keys())}'
        )


def glob(path: path_type, storage_options: Dict = None) -> List[path_type]:
    if "*" in path:
        fs = get_fs(path, storage_options)
        return fs.glob(path)
    else:
        return [path]


def file_size(path: path_type, storage_options: Dict = None) -> int:
    fs = get_fs(path, storage_options)
    return fs.stat(path)["size"]


def open_file(
    path: path_type,
    mode: str = "rb",
    compression: str = None,
    storage_options: Dict = None,
):
    fs = get_fs(path, storage_options)
    file = fs.open(path, mode=mode)

    if compression is not None:
        file = compress(file, compression)

    return file



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import functools
import logging
import os
from collections import defaultdict
from typing import Any, Dict, Optional

import xoscar as mo
import yaml

from ...constants import MARS_PROFILING_RESULTS_DIR
from ...core.operand import Fetch, FetchShuffle
from ...lib.aio import AioFileObject, Isolation, alru_cache
from ...lib.filesystem import get_fs, get_scheme, open_file
from ...services.subtask import Subtask, SubtaskGraph
from ...services.task.core import Task
from ...typing import ChunkType, TileableType
from ...utils import calc_data_size

logger = logging.getLogger(__name__)


def collect_on_demand(f):
    @functools.wraps(f)
    async def wrapper(self, *args, **kwargs):
        if await self.collect_task_info():
            await f(self, *args, **kwargs)
        else:
            pass

    return wrapper


class TaskInfoCollector:
    def __init__(
        self,
        session_id: str,
        task_id: str,
        address: str,
        collect_task_info: Optional[bool] = None,
    ):
        self._session_id = session_id
        self._task_id = task_id
        self._address = address
        self._result_chunk_to_subtask = dict()
        self._save_dir = os.path.join(self._session_id, self._task_id)
        self._collect_task_info = collect_task_info
        self._subtask_infos = defaultdict(dict)

    async def collect_task_info(self):
        if self._collect_task_info is None:
            task_api = await self._get_task_api()
            return await task_api.collect_task_info_enabled()
        else:
            return self._collect_task_info

    @collect_on_demand
    async def collect_subtask_operand_structure(
        self,
        task: Task,
        subtask_graph: SubtaskGraph,
        stage_id: str,
    ):
        """
        This method may be called once for each stage ID.
        """
        self._validate_task(task)

        if self._task_id not in self._result_chunk_to_subtask:
            self._result_chunk_to_subtask[self._task_id] = dict()

        op_info = {}
        op_save_path = os.path.join(self._save_dir, f"{stage_id}_stage_operand.yaml")
        subtask_info = {}
        subtask_save_path = os.path.join(
            self._save_dir, f"{stage_id}_stage_subtask.yaml"
        )
        for subtask in subtask_graph.topological_iter():
            chunk_graph = subtask.chunk_graph
            for c in chunk_graph.results:
                self._result_chunk_to_subtask[self._task_id][c.key] = subtask.subtask_id

            visited = set()
            subtask_info = dict()
            subtask_info["pre_subtasks"] = list()
            subtask_info["ops"] = list()
            subtask_info["stage_id"] = stage_id

            for node in chunk_graph.iter_nodes():
                op = node.op
                if isinstance(op, (Fetch, FetchShuffle)):  # pragma: no cover
                    continue
                if op.key in visited:  # pragma: no cover
                    continue

                subtask_info["ops"].append(op.key)
                op_info = dict()
                op_name = type(op).__name__
                op_info["op_name"] = op_name
                if op.stage is not None:  # pragma: no cover
                    op_info["stage"] = op.stage.name
                else:
                    op_info["stage"] = None
                op_info["subtask_id"] = subtask.subtask_id
                op_info["stage_id"] = stage_id
                op_info["predecessors"] = list()
                op_info["inputs"] = dict()

                for input_chunk in op.inputs or []:
                    if input_chunk.key not in visited:
                        op_info["predecessors"].append(input_chunk.key)
                        if (
                            isinstance(input_chunk.op, (Fetch, FetchShuffle))
                            and input_chunk.key
                            in self._result_chunk_to_subtask[self._task_id]
                        ):  # pragma: no cover
                            subtask_info["pre_subtasks"].append(
                                self._result_chunk_to_subtask[self._task_id][
                                    input_chunk.key
                                ]
                            )

                        chunk_dict = dict()
                        if (
                            hasattr(input_chunk, "dtypes")
                            and input_chunk.dtypes is not None
                        ):  # pragma: no cover
                            unique_dtypes = list(set(input_chunk.dtypes.map(str)))
                            chunk_dict["dtypes"] = unique_dtypes
                        elif (
                            hasattr(input_chunk, "dtype")
                            and input_chunk.dtype is not None
                        ):
                            chunk_dict["dtype"] = str(input_chunk.dtype)

                        op_info["inputs"][input_chunk.key] = chunk_dict

                op_info[op.key] = op_info
                visited.add(op.key)

            subtask_info[subtask.subtask_id] = subtask_info

        await self._save_task_info(op_info, op_save_path)
        await self._save_task_info(subtask_info, subtask_save_path)

    @collect_on_demand
    async def collect_result_nodes(
        self, task: Task, subtask_graphs: list[SubtaskGraph]
    ):
        """
        This method will only be called once for each task.
        """
        # TODO: for each stage, would it be better to have an individual file?
        self._validate_task(task)

        result_op_keys = []
        result_subtask_keys = []
        subtask_graph = subtask_graphs[-1]
        for subtask in subtask_graph.topological_iter():
            if len(subtask_graph._successors[subtask]) == 0:
                result_subtask_keys.append(subtask.subtask_id)
                for result in subtask.chunk_graph.results:
                    result_op_keys.append(result.op.key)

        save_path = os.path.join(self._save_dir, "result_nodes.yaml")
        asyncio.create_task(
            self._save_task_info(
                {"op": result_op_keys, "subtask": result_subtask_keys},
                save_path,
            )
        )

    @collect_on_demand
    async def collect_tileable_structure(
        self, task: Task, tileable_to_subtasks: dict[TileableType, list[Subtask]]
    ):
        """
        This method will only be called once for each task.
        """
        self._validate_task(task)
        tileable_dict = dict()
        for tileable, subtasks in tileable_to_subtasks.items():
            tileable_dict[tileable.key] = dict()
            tileable_dict[tileable.key]["subtasks"] = [x.subtask_id for x in subtasks]
            tileable_dict[tileable.key]["name"] = type(tileable.op).__name__
            tileable_dict[tileable.key]["pre_tileables"] = [
                x.key for x in tileable.op.inputs
            ]

        save_path = os.path.join(self._save_dir, "tileables.yaml")
        await self._save_task_info(tileable_dict, save_path)

    @collect_on_demand
    async def append_runtime_subtask_info(
        self,
        subtask: Subtask,
        band: tuple,
        slot_id: int,
        stored_keys: list[str],
        store_sizes: dict[str, int],
        memory_sizes: dict[str, int],
        cost_times: dict[str, tuple],
    ):
        """
        This method will be called once for each subtask.
        """
        self._validate_subtask(subtask)

        subtask_dict = dict()
        subtask_dict["band"] = band
        subtask_dict["slot_id"] = slot_id
        subtask_dict.update(cost_times)
        result_chunks_dict = dict()
        stored_keys = list(set(stored_keys))
        for key in stored_keys:
            chunk_dict = dict()
            chunk_dict["memory_size"] = memory_sizes[key]
            chunk_dict["store_size"] = store_sizes[key]
            if isinstance(key, tuple):
                key = key[0]
            result_chunks_dict[key] = chunk_dict
        subtask_dict["result_chunks"] = result_chunks_dict

        self._subtask_infos[subtask.subtask_id].update(subtask_dict)

    @collect_on_demand
    async def append_runtime_operand_info(
        self,
        subtask: Subtask,
        start: float,
        end: float,
        chunk: ChunkType,
        processor_context,
    ):
        """
        This method will be called once for each chunk.
        """
        self._validate_subtask(subtask)

        op = chunk.op
        if isinstance(op, (Fetch, FetchShuffle)):
            return
        op_info = dict()
        op_key = op.key
        op_info["op_name"] = type(op).__name__
        op_info["subtask_id"] = subtask.subtask_id
        op_info["execute_time"] = (start, end)
        if chunk.key in processor_context:
            op_info["memory_use"] = calc_data_size(processor_context[chunk.key])
            op_info["result_count"] = 1
        else:
            cnt = 0
            total_size = 0
            for k, v in processor_context.items():
                if isinstance(k, tuple) and k[0] == chunk.key:
                    cnt += 1
                    total_size += calc_data_size(v)
            op_info["memory_use"] = total_size
            op_info["result_count"] = cnt

        if "op_info" not in self._subtask_infos[subtask.subtask_id]:
            self._subtask_infos[subtask.subtask_id]["op_info"] = dict()
        self._subtask_infos[subtask.subtask_id]["op_info"][op_key] = op_info

    @collect_on_demand
    async def flush_subtask_info(self, subtask_id: str):
        if subtask_id in self._subtask_infos:
            save_path = os.path.join(self._save_dir, f"{subtask_id}_subtask_info.yaml")
            subtask_info = self._subtask_infos[subtask_id]
            del self._subtask_infos[subtask_id]
            await self._save_task_info(subtask_info, save_path)

    @collect_on_demand
    async def collect_fetch_time(
        self, subtask: Subtask, fetch_start: float, fetch_end: float
    ):
        # TODO: merge into subtask info so that the number of files could be reduced greatly.
        """
        This method will be called once on worker for each subtask.
        """
        self._validate_subtask(subtask)

        save_path = os.path.join(
            self._save_dir, f"{subtask.subtask_id}_subtask_fetch_time.yaml"
        )
        asyncio.create_task(
            self._save_task_info(
                {subtask.subtask_id: (fetch_start, fetch_end)},
                save_path,
            )
        )

    @alru_cache
    async def _get_task_api(self):
        from .api.oscar import TaskAPI

        return await TaskAPI.create(
            session_id=self._session_id, local_address=self._address
        )

    def _validate_task(self, task: Task):
        assert task.session_id == self._session_id
        assert task.task_id == self._task_id

    def _validate_subtask(self, subtask: Subtask):
        assert subtask.session_id == self._session_id
        assert subtask.task_id == self._task_id

    async def _save_task_info(self, task_info: dict, path: str):
        task_api = await self._get_task_api()
        asyncio.create_task(task_api.save_task_info(task_info, path))


class TaskInfoCollectorActor(mo.Actor):
    def __init__(self, profiling_config: Optional[Dict[str, Any]] = None):
        super().__init__()
        if profiling_config is None:
            profiling_config = dict()
        experimental_profiling_config = profiling_config.get("experimental", dict())
        self._collect_task_info_enabled = experimental_profiling_config.get(
            "collect_task_info_enabled", False
        )
        if self._collect_task_info_enabled:
            self._task_info_root_path = experimental_profiling_config.get(
                "task_info_root_path", None
            )
            if self._task_info_root_path is None:
                self._task_info_root_path = (
                    self._get_or_create_default_profiling_results_dir()
                )
            logger.info(f"Task info root path: {self._task_info_root_path}")
            self._task_info_storage_options = experimental_profiling_config.get(
                "task_info_storage_options", {}
            )

            self._scheme = get_scheme(self._task_info_root_path)
            self._fs = get_fs(
                self._task_info_root_path, self._task_info_storage_options
            )
            self._loop = asyncio.new_event_loop()
            self._isolation = Isolation(self._loop)
            self._isolation.start()

    @staticmethod
    def _get_or_create_default_profiling_results_dir():
        os.makedirs(MARS_PROFILING_RESULTS_DIR, exist_ok=True)
        return MARS_PROFILING_RESULTS_DIR

    async def __pre_destroy__(self):
        if self._collect_task_info_enabled:
            self._isolation.stop()

    async def collect_task_info_enabled(self):
        return self._collect_task_info_enabled

    async def save_task_info(self, info: dict, path: str):
        async def save_info():
            try:
                abs_save_path = os.path.join(self._task_info_root_path, path)
                abs_save_dir, _ = os.path.split(abs_save_path)

                # mkdir when saving to a local file.
                if self._scheme == "file" and not self._fs.exists(abs_save_dir):
                    self._fs.mkdir(abs_save_dir, create_parents=True)

                # since append is not available for object storages, create a new file every time.
                async with AioFileObject(open_file(abs_save_path, "w")) as f:
                    s = await asyncio.create_task(asyncio.to_thread(yaml.dump, info))
                    await f.write(s)
            except:  # noqa: E722   # pragma: no cover
                logger.error(f"Failed to save task info", exc_info=True)

        asyncio.run_coroutine_threadsafe(save_info(), self._loop)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas

from ..core.adapter import (
    MarsEWM,
    MarsExpanding,
    MarsGetAttrProxy,
    MarsRolling,
    register_converter,
)
from ..core.utils.docstring import attach_module_callable_docstring
from .mars_adapters.core import install_members


@register_converter(from_cls_list=[MarsRolling])
class Rolling(MarsGetAttrProxy):
    pass


install_members(
    Rolling,
    MarsRolling,
    pandas,
    pandas.core.window.rolling.Rolling,
)
attach_module_callable_docstring(Rolling, pandas, pandas.core.window.rolling.Rolling)


@register_converter(from_cls_list=[MarsExpanding])
class Expanding(MarsGetAttrProxy):
    pass


install_members(
    Expanding,
    MarsExpanding,
    pandas,
    pandas.core.window.expanding.Expanding,
)
attach_module_callable_docstring(
    Expanding, pandas, pandas.core.window.expanding.Expanding
)


@register_converter(from_cls_list=[MarsEWM])
class ExponentialMovingWindow(MarsGetAttrProxy):
    pass


install_members(
    ExponentialMovingWindow,
    MarsEWM,
    pandas,
    pandas.core.window.ewm.ExponentialMovingWindow,
)
attach_module_callable_docstring(
    ExponentialMovingWindow, pandas, pandas.core.window.ewm.ExponentialMovingWindow
)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import mmap
import os
import warnings
from typing import Any, Dict, List, Tuple, Union

import cloudpickle
from xoscar.serialization import deserialize
from xoscar.serialization.aio import BUFFER_SIZES_NAME, get_header_length

from ..utils import implements
from .base import ObjectInfo, StorageBackend, StorageLevel, register_storage_backend
from .core import BufferWrappedFileObject, StorageFileObject
from .filesystem import DiskStorage


class MMAPFileObject(BufferWrappedFileObject):
    def __init__(self, object_id: Any, view: memoryview, size: int):
        super().__init__(object_id, "r", size=size)
        self._mv = view
        self._buffer = view

    @property
    def object_id(self):
        return self._object_id

    @property
    def buffer(self):
        return self._buffer

    def _write_init(self):  # pragma: no cover
        raise NotImplementedError("MMAPFileObject only supports read operations.")

    def write(self, content: Union[bytes, memoryview]):  # pragma: no cover
        raise NotImplementedError("MMAPFileObject only supports read operations.")

    def _read_init(self):  # pragma: no cover
        pass

    def read(self, size=-1):
        offset = self._offset
        size = self._size if size < 0 else size
        end = min(self._size, offset + size)
        result = self._mv[offset:end]
        self._offset = end
        return result

    def _read_close(self):  # pragma: no cover
        pass

    def _write_close(self):  # pragma: no cover
        pass

    def seek(self, offset: int, whence: int = os.SEEK_SET):
        return self._get_new_offset(offset, whence=whence)


class MMAPStorageFileObject(StorageFileObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@register_storage_backend
class MMAPStorage(DiskStorage):
    name = "mmap"
    is_seekable = True

    def __init__(self, **kw):
        DiskStorage.__init__(self, **kw)

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        disk_init, disk_teardown = await DiskStorage.setup(**kwargs)

        return disk_init, disk_teardown

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        fs = kwargs.get("fs")
        root_dirs = kwargs.get("root_dirs")
        try:
            for d in root_dirs:
                fs.delete(d, recursive=True)
        except:  # pragma: no cover
            warnings.warn(
                f"Unexpected exceptions occur when deleting directories, "
                f"please delete these directories manually: {root_dirs}"
            )

    @property
    @implements(StorageBackend.level)
    def level(self) -> StorageLevel:
        return StorageLevel.MEMORY

    @property
    @implements(StorageBackend.size)
    def size(self) -> Union[int, None]:
        return self._size

    @staticmethod
    def _get_file_via_mmap(object_id: str):
        with open(object_id, "r+b") as f:
            mm = mmap.mmap(f.fileno(), 0)
            v = memoryview(mm)
            header_bytes = v[:11]
            header_length = get_header_length(header_bytes)
            header = cloudpickle.loads(v[11 : 11 + header_length])
            buffer_sizes = header[0].pop(BUFFER_SIZES_NAME)
            # get buffers
            buffers = []
            start = 11 + header_length
            for size in buffer_sizes:
                buffers.append(v[start : start + size])
                start = start + size
            return deserialize(header, buffers)

    @implements(StorageBackend.get)
    async def get(self, object_id, **kwargs) -> object:
        return await asyncio.to_thread(self._get_file_via_mmap, object_id)

    @implements(StorageBackend.put)
    async def put(self, obj, importance=0) -> ObjectInfo:
        return await super().put(obj, importance=importance)

    @implements(StorageBackend.object_info)
    async def object_info(self, object_id) -> ObjectInfo:
        return await super().object_info(object_id)

    @implements(StorageBackend.delete)
    async def delete(self, object_id):
        await super().delete(object_id)

    @implements(StorageBackend.open_writer)
    async def open_writer(self, size=None) -> StorageFileObject:
        return await super().open_writer(size=size)

    @implements(StorageBackend.open_reader)
    async def open_reader(self, object_id) -> StorageFileObject:
        with open(object_id, "r+b") as f:
            mm = mmap.mmap(f.fileno(), 0)
            file = MMAPFileObject(object_id, memoryview(mm), mm.size())
            return MMAPStorageFileObject(file, object_id=object_id)

    @implements(StorageBackend.list)
    async def list(self) -> List:  # pragma: no cover
        raise NotImplementedError("MMAP storage does not support list")



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.ensemble as sk_en

from ...._mars.learn import ensemble as mars_en
from ...._mars.learn.ensemble import BaggingClassifier as MarsBaggingClassifier
from ...._mars.learn.ensemble import BaggingRegressor as MarsBaggingRegressor
from ...._mars.learn.ensemble import IsolationForest as MarsIsolationForest
from ....core.utils.docstring import attach_module_callable_docstring
from ...utils import SKLearnBase, _collect_module_callables, _install_cls_members


class BaggingClassifier(SKLearnBase):
    _marscls = MarsBaggingClassifier


class BaggingRegressor(SKLearnBase):
    _marscls = MarsBaggingRegressor


class IsolationForest(SKLearnBase):
    _marscls = MarsIsolationForest


SKLEARN_EN_CLS_MAP = {
    BaggingClassifier: MarsBaggingClassifier,
    IsolationForest: MarsIsolationForest,
    BaggingRegressor: MarsBaggingRegressor,
}

MARS_SKLEARN_EN_CALLABLES = _collect_module_callables(
    mars_en, sk_en, skip_members=["register_op"]
)
_install_cls_members(SKLEARN_EN_CLS_MAP, MARS_SKLEARN_EN_CALLABLES, sk_en)
attach_module_callable_docstring(BaggingClassifier, sk_en, sk_en.BaggingClassifier)
attach_module_callable_docstring(BaggingRegressor, sk_en, sk_en.BaggingRegressor)
attach_module_callable_docstring(IsolationForest, sk_en, sk_en.IsolationForest)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import _version
from .config import options
from .core.context import get_context
from .core.entrypoints import init_extension_entrypoints
from .session import execute, fetch, fetch_log, new_session, stop_server

# load third party extensions.
init_extension_entrypoints()
del init_extension_entrypoints

__version__ = _version.get_versions()["version"]



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import sys
from typing import Dict, List, Optional, Union

import numpy as np
import xoscar as mo
from xoscar.backends.router import Router
from xoscar.metrics import init_metrics

from ...lib.aio import get_isolation
from ...resource import cpu_count, cuda_count, mem_total
from ...services import NodeRole
from ...services.task.execution.api import ExecutionConfig
from ...typing import ClientType, ClusterType
from ..utils import get_third_party_modules_from_config, load_config
from .pool import create_supervisor_actor_pool, create_worker_actor_pool
from .service import start_supervisor, start_worker, stop_supervisor, stop_worker
from .session import AbstractSession, _new_session, ensure_isolation_created

logger = logging.getLogger(__name__)

# The default config file.
DEFAULT_CONFIG_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "config.yml"
)

# the default times to retry subtask.
DEFAULT_SUBTASK_MAX_RETRIES = 3
# the default time to cancel a subtask.
DEFAULT_SUBTASK_CANCEL_TIMEOUT = 5


def _load_config(config: Union[str, Dict] = None):
    return load_config(config, default_config_file=DEFAULT_CONFIG_FILE)


async def new_cluster_in_isolation(
    address: str = "0.0.0.0",
    n_worker: int = 1,
    n_cpu: Union[int, str] = "auto",
    mem_bytes: Union[int, str] = "auto",
    cuda_devices: Union[List[int], str] = "auto",
    subprocess_start_method: str = None,
    backend: str = None,
    config: Union[str, Dict] = None,
    web: bool = True,
    timeout: float = None,
    n_supervisor_process: int = 0,
    numa_external_addr_scheme: str = None,
    numa_enable_internal_addr: bool = None,
    gpu_external_addr_scheme: str = None,
    gpu_enable_internal_addr: bool = None,
    io_external_addr_scheme: Optional[str] = None,
    io_enable_internal_addr: Optional[bool] = None,
    external_addr_scheme: str = None,
    enable_internal_addr: bool = None,
    oscar_extra_conf: dict = None,
    log_config: dict = None,
    storage_config: Optional[dict] = None,
) -> ClientType:
    cluster = LocalCluster(
        address,
        n_worker,
        n_cpu,
        mem_bytes,
        cuda_devices,
        subprocess_start_method,
        backend,
        config,
        web,
        n_supervisor_process,
        numa_external_addr_scheme=numa_external_addr_scheme,
        numa_enable_internal_addr=numa_enable_internal_addr,
        gpu_external_addr_scheme=gpu_external_addr_scheme,
        gpu_enable_internal_addr=gpu_enable_internal_addr,
        io_external_addr_scheme=io_external_addr_scheme,
        io_enable_internal_addr=io_enable_internal_addr,
        external_addr_scheme=external_addr_scheme,
        enable_internal_addr=enable_internal_addr,
        oscar_extra_conf=oscar_extra_conf,
        log_config=log_config,
        storage_config=storage_config,
    )
    await cluster.start()
    return await LocalClient.create(cluster, timeout)


async def new_cluster(
    address: str = "0.0.0.0",
    n_worker: int = 1,
    n_cpu: Union[int, str] = "auto",
    mem_bytes: Union[int, str] = "auto",
    cuda_devices: Union[List[int], str] = "auto",
    subprocess_start_method: str = None,
    backend: str = None,
    config: Union[str, Dict] = None,
    web: bool = True,
    loop: asyncio.AbstractEventLoop = None,
    use_uvloop: Union[bool, str] = "auto",
    n_supervisor_process: int = 0,
    numa_external_addr_scheme: str = None,
    numa_enable_internal_addr: bool = None,
    gpu_external_addr_scheme: str = None,
    gpu_enable_internal_addr: bool = None,
    io_external_addr_scheme: Optional[str] = None,
    io_enable_internal_addr: Optional[bool] = None,
    external_addr_scheme: str = None,
    enable_internal_addr: bool = None,
    oscar_extra_conf: dict = None,
) -> ClientType:
    coro = new_cluster_in_isolation(
        address,
        n_worker=n_worker,
        n_cpu=n_cpu,
        mem_bytes=mem_bytes,
        cuda_devices=cuda_devices,
        subprocess_start_method=subprocess_start_method,
        backend=backend,
        config=config,
        web=web,
        n_supervisor_process=n_supervisor_process,
        numa_external_addr_scheme=numa_external_addr_scheme,
        numa_enable_internal_addr=numa_enable_internal_addr,
        gpu_external_addr_scheme=gpu_external_addr_scheme,
        gpu_enable_internal_addr=gpu_enable_internal_addr,
        io_external_addr_scheme=io_external_addr_scheme,
        io_enable_internal_addr=io_enable_internal_addr,
        external_addr_scheme=external_addr_scheme,
        enable_internal_addr=enable_internal_addr,
        oscar_extra_conf=oscar_extra_conf,
    )
    isolation = ensure_isolation_created(dict(loop=loop, use_uvloop=use_uvloop))
    fut = asyncio.run_coroutine_threadsafe(coro, isolation.loop)
    client = await asyncio.wrap_future(fut)
    client.session.as_default()
    return client


async def stop_cluster(cluster: ClusterType):
    isolation = get_isolation()
    coro = cluster.stop()
    await asyncio.wrap_future(asyncio.run_coroutine_threadsafe(coro, isolation.loop))
    Router.set_instance(None)


class LocalCluster:
    def __init__(
        self: ClusterType,
        address: str = "0.0.0.0",
        n_worker: int = 1,
        n_cpu: Union[int, str] = "auto",
        mem_bytes: Union[int, str] = "auto",
        cuda_devices: Union[List[int], List[List[int]], str] = "auto",
        subprocess_start_method: str = None,
        backend: str = None,
        config: Union[str, Dict] = None,
        web: Union[bool, str] = "auto",
        n_supervisor_process: int = 0,
        numa_external_addr_scheme: str = None,
        numa_enable_internal_addr: bool = None,
        gpu_external_addr_scheme: str = None,
        gpu_enable_internal_addr: bool = None,
        io_external_addr_scheme: Optional[str] = None,
        io_enable_internal_addr: Optional[bool] = None,
        external_addr_scheme: str = None,
        enable_internal_addr: str = None,
        oscar_extra_conf: dict = None,
        log_config: dict = None,
        storage_config: Optional[dict] = None,
    ):
        # auto choose the subprocess_start_method.
        if subprocess_start_method is None:
            subprocess_start_method = (
                "spawn" if sys.platform == "win32" else "forkserver"
            )
        self._address = address
        self._n_worker = n_worker
        self._n_cpu = cpu_count() if n_cpu == "auto" else n_cpu
        self._mem_bytes = mem_total() if mem_bytes == "auto" else mem_bytes
        self._cuda_devices = self._get_cuda_devices(cuda_devices, n_worker)
        self._subprocess_start_method = subprocess_start_method
        self._config = load_config(config, default_config_file=DEFAULT_CONFIG_FILE)
        execution_config = ExecutionConfig.from_config(self._config, backend=backend)
        self._log_config = log_config
        self._backend = execution_config.backend
        self._web = web
        self._n_supervisor_process = n_supervisor_process

        execution_config.merge_from(
            ExecutionConfig.from_params(
                backend=self._backend,
                n_worker=self._n_worker,
                n_cpu=self._n_cpu,
                mem_bytes=self._mem_bytes,
                cuda_devices=self._cuda_devices,
                subtask_cancel_timeout=self._config.get("scheduling", {}).get(
                    "subtask_cancel_timeout", DEFAULT_SUBTASK_CANCEL_TIMEOUT
                ),
                subtask_max_retries=self._config.get("scheduling", {}).get(
                    "subtask_max_retries", DEFAULT_SUBTASK_MAX_RETRIES
                ),
            )
        )

        # process oscar config
        self._process_oscar_config(
            numa_external_addr_scheme=numa_external_addr_scheme,
            numa_enable_internal_addr=numa_enable_internal_addr,
            gpu_external_addr_scheme=gpu_external_addr_scheme,
            gpu_enable_internal_addr=gpu_enable_internal_addr,
            io_external_addr_scheme=io_external_addr_scheme,
            io_enable_internal_addr=io_enable_internal_addr,
            external_addr_scheme=external_addr_scheme,
            enable_internal_addr=enable_internal_addr,
            oscar_extra_conf=oscar_extra_conf,
        )

        # make memory allocation policy more aggressive on local by
        # assuming all the memory is available.
        if self._config.get("scheduling", {}).get("mem_hard_limit", None) is None:
            self._config["scheduling"]["mem_hard_limit"] = None

        if storage_config is not None:
            backend = list(storage_config.keys())[0]
            self._config["storage"]["backends"] = [backend]
            self._config["storage"][backend] = storage_config[backend]

        self._bands_to_resource = execution_config.get_deploy_band_resources()
        self._supervisor_pool = None
        self._worker_pools = []

        self.supervisor_address = None
        self.web_address = None

    def _process_oscar_config(
        self,
        numa_external_addr_scheme: str = None,
        numa_enable_internal_addr: bool = None,
        gpu_external_addr_scheme: str = None,
        gpu_enable_internal_addr: bool = None,
        io_external_addr_scheme: Optional[str] = None,
        io_enable_internal_addr: Optional[bool] = None,
        external_addr_scheme: str = None,
        enable_internal_addr: str = None,
        oscar_extra_conf: dict = None,
    ):
        # process oscar config
        assert "oscar" in self._config
        oscar_config = self._config["oscar"]
        numa_config = oscar_config["numa"]
        numa_external_addr_scheme = (
            numa_external_addr_scheme
            if numa_external_addr_scheme is not None
            else external_addr_scheme
        )
        if numa_external_addr_scheme:
            numa_config["external_addr_scheme"] = numa_external_addr_scheme
        numa_enable_internal_addr = (
            numa_enable_internal_addr
            if numa_enable_internal_addr is not None
            else enable_internal_addr
        )
        if numa_enable_internal_addr is not None:
            numa_config["enable_internal_addr"] = numa_enable_internal_addr
        gpu_config = oscar_config["gpu"]
        gpu_external_addr_scheme = (
            gpu_external_addr_scheme
            if gpu_external_addr_scheme is not None
            else external_addr_scheme
        )
        if gpu_external_addr_scheme:
            gpu_config["external_addr_scheme"] = gpu_external_addr_scheme
        gpu_enable_internal_addr = (
            gpu_enable_internal_addr
            if gpu_enable_internal_addr is not None
            else enable_internal_addr
        )
        if gpu_enable_internal_addr is not None:
            gpu_config["enable_internal_addr"] = gpu_enable_internal_addr

        io_config = oscar_config["io"]
        io_external_addr_scheme = (
            io_external_addr_scheme
            if io_external_addr_scheme is not None
            else external_addr_scheme
        )
        if io_external_addr_scheme is not None:
            io_config["external_addr_scheme"] = io_external_addr_scheme
        io_enable_internal_addr = (
            io_enable_internal_addr
            if io_enable_internal_addr is not None
            else enable_internal_addr
        )
        if io_enable_internal_addr is not None:
            io_config["enable_internal_addr"] = io_enable_internal_addr

        if oscar_extra_conf is not None:
            oscar_config["extra_conf"] = oscar_extra_conf

    @staticmethod
    def _get_cuda_devices(cuda_devices, n_worker):
        if cuda_devices == "auto":
            total = cuda_count()
            all_devices = np.arange(total)
            return [list(arr) for arr in np.array_split(all_devices, n_worker)]

        else:
            if not cuda_devices:
                return [[]] * n_worker
            elif isinstance(cuda_devices[0], int):
                assert n_worker == 1
                return [cuda_devices]
            else:
                assert len(cuda_devices) == n_worker
                return cuda_devices

    @property
    def backend(self):
        return self._backend

    @property
    def external_address(self):
        return self._supervisor_pool.external_address

    async def start(self):
        await self._start_supervisor_pool()
        await self._start_worker_pools()
        # start service
        await self._start_service()

        # init metrics to guarantee metrics use in driver
        metric_configs = self._config.get("metrics", {})
        metric_backend = metric_configs.get("backend")
        init_metrics(metric_backend, config=metric_configs.get(metric_backend))

        if self._web:
            from ...services.web.supervisor import WebActor

            web_actor = await mo.actor_ref(
                WebActor.default_uid(), address=self.supervisor_address
            )
            self.web_address = await web_actor.get_web_address()
            logger.warning("Web service started at %s", self.web_address)

    async def _start_supervisor_pool(self):
        supervisor_modules = get_third_party_modules_from_config(
            self._config, NodeRole.SUPERVISOR
        )
        self._supervisor_pool = await create_supervisor_actor_pool(
            self._address,
            n_process=self._n_supervisor_process,
            modules=supervisor_modules,
            subprocess_start_method=self._subprocess_start_method,
            metrics=self._config.get("metrics", {}),
            web=self._web,
            # passing logging conf to config logging when create pools
            logging_conf=self._log_config,
            oscar_config=self._config.get("oscar"),
        )
        self.supervisor_address = self._supervisor_pool.external_address

    async def _start_worker_pools(self):
        worker_modules = get_third_party_modules_from_config(
            self._config, NodeRole.WORKER
        )
        for band_to_resource, worker_devices in zip(
            self._bands_to_resource, self._cuda_devices
        ):
            worker_pool = await create_worker_actor_pool(
                self._address,
                band_to_resource,
                modules=worker_modules,
                subprocess_start_method=self._subprocess_start_method,
                metrics=self._config.get("metrics", {}),
                cuda_devices=worker_devices,
                web=self._web,
                # passing logging conf to config logging when create pools
                logging_conf=self._log_config,
                oscar_config=self._config.get("oscar"),
            )
            self._worker_pools.append(worker_pool)

    async def _start_service(self):
        self._web = await start_supervisor(
            self.supervisor_address, config=self._config, web=self._web
        )
        for worker_pool, band_to_resource in zip(
            self._worker_pools, self._bands_to_resource
        ):
            await start_worker(
                worker_pool.external_address,
                self.supervisor_address,
                band_to_resource,
                config=self._config,
            )

    async def stop(self):
        from .session import SessionAPI

        # delete all sessions
        session_api = await SessionAPI.create(self._supervisor_pool.external_address)
        await session_api.delete_all_sessions()

        for worker_pool in self._worker_pools:
            await stop_worker(worker_pool.external_address, self._config)
        await stop_supervisor(self._supervisor_pool.external_address, self._config)
        for worker_pool in self._worker_pools:
            await worker_pool.stop()
        await self._supervisor_pool.stop()
        AbstractSession.reset_default()
        Router.set_instance(None)


class LocalClient:
    def __init__(self: ClientType, cluster: ClusterType, session: AbstractSession):
        self._cluster = cluster
        self.session = session

    @classmethod
    async def create(
        cls,
        cluster: LocalCluster,
        timeout: float = None,
    ) -> ClientType:
        session = await _new_session(
            cluster.external_address,
            backend=cluster.backend,
            default=True,
            timeout=timeout,
        )
        client = LocalClient(cluster, session)
        session.client = client
        return client

    @property
    def web_address(self):
        return self._cluster.web_address

    async def __aenter__(self):
        return self

    async def __aexit__(self, *_):
        await self.stop()

    async def stop(self):
        await stop_cluster(self._cluster)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import tensor as mt


def rankdata(a, method="average", *, axis=None):
    """Assign ranks to data, dealing with ties appropriately.
    By default (``axis=None``), the data array is first flattened, and a flat
    array of ranks is returned. Separately reshape the rank array to the
    shape of the data array if desired (see Examples).
    Ranks begin at 1.  The `method` argument controls how ranks are assigned
    to equal values.  See [1]_ for further discussion of ranking methods.
    Parameters
    ----------
    a : array_like
        The array of values to be ranked.
    method : {'average', 'min', 'max', 'dense', 'ordinal'}, optional
        The method used to assign ranks to tied elements.
        The following methods are available (default is 'average'):
          * 'average': The average of the ranks that would have been assigned to
            all the tied values is assigned to each value.
          * 'min': The minimum of the ranks that would have been assigned to all
            the tied values is assigned to each value.  (This is also
            referred to as "competition" ranking.)
          * 'max': The maximum of the ranks that would have been assigned to all
            the tied values is assigned to each value.
          * 'dense': Like 'min', but the rank of the next highest element is
            assigned the rank immediately after those assigned to the tied
            elements.
          * 'ordinal': All values are given a distinct rank, corresponding to
            the order that the values occur in `a`.
    axis : {None, int}, optional
        Axis along which to perform the ranking. If ``None``, the data array
        is first flattened.
    Returns
    -------
    ranks : ndarray
         An array of size equal to the size of `a`, containing rank
         scores.
    References
    ----------
    .. [1] "Ranking", https://en.wikipedia.org/wiki/Ranking
    Examples
    --------
    >>> from mars.tensor.stats import rankdata
    >>> rankdata([0, 2, 3, 2]).execute()
    array([ 1. ,  2.5,  4. ,  2.5])
    >>> rankdata([0, 2, 3, 2], method='min').execute()
    array([ 1,  2,  4,  2])
    >>> rankdata([0, 2, 3, 2], method='max').execute()
    array([ 1,  3,  4,  3])
    >>> rankdata([0, 2, 3, 2], method='dense').execute()
    array([ 1,  2,  3,  2])
    >>> rankdata([0, 2, 3, 2], method='ordinal').execute()
    array([ 1,  2,  4,  3])
    >>> rankdata([[0, 2], [3, 2]]).reshape(2,2).execute()
    array([[1. , 2.5],
          [4. , 2.5]])
    >>> rankdata([[0, 2, 2], [3, 2, 5]], axis=1).execute()
    array([[1. , 2.5, 2.5],
           [2. , 1. , 3. ]])
    """
    if method not in ("average", "min", "max", "dense", "ordinal"):
        raise ValueError('unknown method "{0}"'.format(method))

    if axis is not None:
        a = np.asarray(a)
        if a.size == 0:
            np.core.multiarray.normalize_axis_index(axis, a.ndim)
            dt = np.float64 if method == "average" else np.int_
            return mt.empty(a.shape, dtype=dt)
        return mt.tensor(np.apply_along_axis(rankdata, axis, a, method))

    arr = mt.ravel(mt.asarray(a))
    algo = "mergesort" if method == "ordinal" else "quicksort"
    sorter = mt.argsort(arr, kind=algo)

    inv = mt.empty(sorter.size, dtype=np.intp)
    inv[sorter] = mt.arange(sorter.size, dtype=np.intp)

    if method == "ordinal":
        return inv + 1

    arr = arr[sorter]
    obs = mt.r_[True, arr[1:] != arr[:-1]]
    dense = obs.cumsum()[inv]

    if method == "dense":
        return dense

    count = mt.r_[mt.nonzero(obs)[0], len(obs)]

    if method == "max":
        return count[dense]

    if method == "min":
        return count[dense - 1] + 1

    return 0.5 * (count[dense] + count[dense - 1] + 1)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes
from ...serialization.serializables import (
    FieldTypes,
    Int32Field,
    StringField,
    TupleField,
)
from . import FetchShuffle, ShuffleFetchType
from .base import Operand, OperandStage, VirtualOperand


class ShuffleProxy(VirtualOperand):
    _op_type_ = opcodes.SHUFFLE_PROXY
    n_mappers = Int32Field("n_mappers", default=0)
    # `n_reducers` will be updated in `MapReduceOperand._new_chunks`
    n_reducers = Int32Field("n_reducers", default=0)

    def _new_chunks(self, inputs, kws=None, **kw):
        self.n_mappers = len(inputs)
        return super()._new_chunks(inputs, kws, **kw)


class MapReduceOperand(Operand):
    """
    An operand for shuffle execution which partitions data by the value in each record’s partition key, and
    send the partitioned data from all mappers to all reducers.
    """

    # for reducer
    reducer_index = TupleField("reducer_index", FieldTypes.uint64)
    # Total reducer nums, which also be shuffle blocks for single mapper.
    n_reducers = Int32Field("n_reducers")
    # The reducer ordinal in all reducers. It's different from reducer_index,
    # which might be a tuple.
    # `reducer_ordinal` will be set in `_new_chunks`.
    reducer_ordinal = Int32Field("reducer_ordinal")
    reducer_phase = StringField("reducer_phase", default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.stage == OperandStage.reduce:
            # for reducer, we assign worker at first
            self.scheduling_hint.reassign_worker = True

    def _new_chunks(self, inputs, kws=None, **kw):
        if getattr(self, "reducer_index", None) is None:
            index = None
            if kws:
                index = kws[0].get("index")
            self.reducer_index = index or kw.get("index")
        if self.stage == OperandStage.reduce:
            # Operands such as `TensorIndexSetValue` will have multiple inputs, some won't be ProxyChunk
            proxy_operands = [c.op for c in inputs if isinstance(c.op, ShuffleProxy)]
            if proxy_operands:
                # For create reduce checks with `FetchShuffle`, `proxy_operands` will be empty.
                proxy = proxy_operands[0]
                self.reducer_ordinal = proxy.n_reducers
                proxy.n_reducers += 1
        return super()._new_chunks(inputs, kws, **kw)

    def get_dependent_data_keys(self):
        from .fetch import FetchShuffle

        if self.stage == OperandStage.reduce:
            inputs = self.inputs or ()
            deps = []
            for inp in inputs:
                if isinstance(inp.op, ShuffleProxy):
                    deps.extend(
                        [(chunk.key, self.reducer_index) for chunk in inp.inputs or ()]
                    )
                elif isinstance(inp.op, FetchShuffle):
                    # fetch shuffle by index doesn't store data keys, so it won't run into this function.
                    assert inp.op.shuffle_fetch_type == ShuffleFetchType.FETCH_BY_KEY
                    deps.extend([(k, self.reducer_index) for k in inp.op.source_keys])
                else:
                    deps.append(inp.key)
            return deps
        return super().get_dependent_data_keys()

    def iter_mapper_keys(self, input_id=0):
        # key is mapper chunk key, index is mapper chunk index.
        input_chunk = self.inputs[input_id]
        if isinstance(input_chunk.op, ShuffleProxy):
            keys = [inp.key for inp in input_chunk.inputs]
        else:
            assert isinstance(input_chunk.op, FetchShuffle), input_chunk.op
            if input_chunk.op.shuffle_fetch_type == ShuffleFetchType.FETCH_BY_INDEX:
                # For fetch shuffle by index, all shuffle block of same reducers are
                # identified by their index. chunk key are not needed any more.
                # so just mock key here.
                # keep this in sync with ray executor `execute_subtask`.
                return list(range(input_chunk.op.n_mappers))
            keys = input_chunk.op.source_keys
        return keys

    def iter_mapper_data(self, ctx, input_id=0, pop=False, skip_none=False):
        for key in self.iter_mapper_keys(input_id):
            try:
                if pop:
                    yield ctx.pop((key, self.reducer_index))
                else:
                    yield ctx[key, self.reducer_index]
            except KeyError:
                if not skip_none:  # pragma: no cover
                    raise
                if not pop:
                    ctx[key, self.reducer_index] = None

    def execute(self, ctx, op):
        """The mapper stage must ensure all mapper blocks are inserted into ctx and no blocks
        for some reducers are missing. This is needed by shuffle fetch by index,
        which shuffle block are identified by the  index instead of data keys.
        For operands implementation simplicity, we can sort the `ctx` by key which are (chunk key, reducer index) tuple
        and relax the insert order requirements.
        """



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def _install():
    import os

    from tornado import web

    from .._mars.services.web.handlers import handlers, static_handlers
    from .index_handler import XorbitsIndexHandler
    from .version_handler import XorbitsVersionHandler

    handlers["/"] = XorbitsIndexHandler
    static_handlers[r"[^\?\&]*/static/(.*)"] = (
        web.StaticFileHandler,
        {"path": os.path.join(os.path.dirname(__file__), "ui", "static")},
    )
    handlers["/api/xorbits/version"] = XorbitsVersionHandler



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .chunks import CHUNK_TYPE, Chunk, ChunkData
from .core import ENTITY_TYPE, Entity, EntityData
from .executable import ExecutableTuple, _ExecuteAndFetchMixin
from .fuse import FUSE_CHUNK_TYPE, FuseChunk, FuseChunkData
from .objects import (
    OBJECT_CHUNK_TYPE,
    OBJECT_TYPE,
    Object,
    ObjectChunk,
    ObjectChunkData,
    ObjectData,
)
from .output_types import (
    OutputType,
    get_chunk_types,
    get_fetch_class,
    get_output_types,
    get_tileable_types,
    register_fetch_class,
    register_output_types,
)
from .tileables import (
    TILEABLE_TYPE,
    HasShapeTileable,
    HasShapeTileableData,
    NotSupportTile,
    Tileable,
    TileableData,
    register,
    unregister,
)
from .utils import recursive_tile, tile



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from . import _version
from .core import run
from .deploy import init, shutdown


def _install():
    from .datasets import _install as _install_datasets
    from .experimental import _install as _install_experimental
    from .lightgbm import _install as _install_lightgbm
    from .numpy import _install as _install_numpy
    from .pandas import _install as _install_pandas
    from .sklearn import _install as _install_sklearn
    from .web import _install as _install_web
    from .xgboost import _install as _install_xgboost

    _install_pandas()
    _install_numpy()
    _install_web()
    _install_lightgbm()
    _install_xgboost()
    _install_datasets()
    _install_experimental()
    _install_sklearn()


_install()
del _install


__version__ = _version.get_versions()["version"]

__all__ = ["init", "shutdown"]



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Tuple, Union

from ...core import ChunkType
from ...tensor.fuse import TensorJAXFuseChunk
from ...tensor.fuse.jax import (
    ARITHMETIC_SUPPORT,
    JAX_INSTALLED,
    REDUCTION_SUPPORT,
    TREE_SUPPORT,
)
from .core import REDUCTION, GraphTraversalOptimizer, _Fuse, register_optimizer

logger = logging.getLogger(__name__)


@register_optimizer
class JAXRuntimeOptimizer(GraphTraversalOptimizer):
    engine = "jax"

    def _can_fuse(self, node: ChunkType) -> Union[bool, object]:
        op = node.op
        op_type = type(op)
        if op_type in REDUCTION_SUPPORT:
            if len(op.axis) == 1 or len(op.axis) == node.ndim:
                return REDUCTION
            else:  # pragma: no cover
                return False
        if op_type not in ARITHMETIC_SUPPORT and op_type not in TREE_SUPPORT:
            return False
        return True

    @classmethod
    def is_available(cls) -> bool:
        return JAX_INSTALLED

    def optimize(self) -> Tuple[List[_Fuse], List[ChunkType]]:
        fuses = self._graph_traverse(
            logger, "Refused fusing for jax because the tail node count > 1."
        )

        if fuses == ([], []):  # pragma: no cover
            return [], []

        return self._fuse_nodes(fuses, TensorJAXFuseChunk)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import itertools
import os.path
from typing import Any, Dict, Mapping, Optional, Sequence, Union

import numpy as np

try:
    # For type hint.
    from datasets import (
        DownloadConfig,
        DownloadMode,
        Features,
        Split,
        VerificationMode,
        Version,
    )
except ImportError:
    DownloadConfig = Any
    DownloadMode = Any
    Features = Any
    Split = Any
    VerificationMode = Any
    Version = Any

from ...._mars.core import recursive_tile
from ...._mars.core.context import get_context
from ...._mars.core.entity import OutputType
from ...._mars.serialization.serializables import (
    BoolField,
    DictField,
    ListField,
    StringField,
)
from ....utils import check_signature_compatible, get_non_default_kwargs
from ...dataset import Dataset
from ...operand import DataOperand, DataOperandMixin
from .rechunk import rechunk


class FromHuggingface(DataOperand, DataOperandMixin):
    path = StringField("path")
    hf_kwargs = DictField("hf_kwargs")
    cache_dir = StringField("cache_dir")
    single_data_file = StringField("single_data_file")
    data_files = ListField("data_files")
    auto_rechunk: bool = BoolField("auto_rechunk")

    def __call__(self):
        from datasets import load_dataset_builder
        from datasets.packaged_modules.folder_based_builder.folder_based_builder import (
            FolderBasedBuilder,
        )

        builder_kwargs = self._get_kwargs(load_dataset_builder, self.hf_kwargs)
        builder = load_dataset_builder(self.path, **builder_kwargs)
        assert "cache_dir" in inspect.signature(load_dataset_builder).parameters
        self.cache_dir = builder.cache_dir
        data_files = builder.config.data_files
        # TODO(codingl2k1): support multiple splits
        split = self.hf_kwargs["split"]
        # TODO(codingl2k1): not pass dtypes if no to_dataframe() called.
        if builder.info.features:
            dtypes = builder.info.features.arrow_schema.empty_table().to_pandas().dtypes
        else:
            dtypes = None
        if dtypes is not None and builder.info.splits:
            shape = (
                builder.info.splits[split].num_examples,
                len(dtypes),
            )
        else:
            shape = (np.nan, np.nan)
        # TODO(codingl2k1): check data_files if can be supported
        # e.g. the datasets mariosasko/test_imagefolder_with_metadata has multiple
        # data files, but some of them are meta, so we can't parallel load it.
        if not isinstance(builder, FolderBasedBuilder):
            if data_files and len(data_files[split]) > 1:
                data_files = list(data_files[split])
            else:
                data_files = None
        else:
            data_files = None
        self.data_files = data_files
        return self.new_tileable([], dtypes=dtypes, shape=shape)

    @classmethod
    def _get_kwargs(cls, obj, kwargs):
        sig_builder = inspect.signature(obj)
        return {
            key: kwargs[key] for key in sig_builder.parameters.keys() if key in kwargs
        }

    @classmethod
    def tile(cls, op: "FromHuggingface"):
        assert len(op.inputs) == 0
        out = op.outputs[0]

        data_files = op.data_files
        # Set op.data_files to None, we don't want every chunk op copy this field.
        op.data_files = None

        chunks = []
        if data_files is not None:
            ctx = get_context()
            # TODO(codingl2k1): make expect worker binding stable for cache reuse.
            all_bands = [b for b in ctx.get_worker_bands() if b[1].startswith("numa-")]
            for index, (f, band) in enumerate(
                zip(data_files, itertools.cycle(all_bands))
            ):
                chunk_op = op.copy().reset_key()
                assert f, "Invalid data file from DatasetBuilder."
                chunk_op.single_data_file = f
                chunk_op.expect_band = band
                # The cache dir can't be shared, because there will be a file lock
                # on the cache dir when initializing builder.
                chunk_op.hf_kwargs = dict(
                    op.hf_kwargs,
                    cache_dir=os.path.join(
                        op.cache_dir, f"part_{index}_{len(data_files)}"
                    ),
                )
                chunk_op.cache_dir = None
                c = chunk_op.new_chunk(inputs=[], index=(index, 0))
                chunks.append(c)
        else:
            chunk_op = op.copy().reset_key()
            chunk_op.single_data_file = None
            chunk_op.cache_dir = None
            chunks.append(chunk_op.new_chunk(inputs=[], index=(0, 0)))
            if op.auto_rechunk:
                ctx = get_context()
                cluster_cpu_count = int(ctx.get_total_n_cpu())
                if (
                    out.shape
                    and not np.isnan(out.shape[0])
                    and out.shape[0] >= cluster_cpu_count
                ):
                    inp = op.copy().new_tileable(
                        op.inputs,
                        chunks=chunks,
                        nsplits=((np.nan,) * len(chunks), (np.nan,)),
                        **out.params,
                    )
                    auto_rechunked = yield from recursive_tile(
                        rechunk(inp, cluster_cpu_count)
                    )
                    return auto_rechunked

        return op.copy().new_tileable(
            op.inputs,
            chunks=chunks,
            nsplits=((np.nan,) * len(chunks), (np.nan,)),
            **out.params,
        )

    @classmethod
    def execute(cls, ctx, op: "FromHuggingface"):
        from datasets import DatasetBuilder, VerificationMode, load_dataset_builder

        # load_dataset_builder from every worker may be slow, but it's error to
        # deserialized a builder instance in a clean process / node, e.g. raise
        # ModuleNotFoundError: No module named 'datasets_modules'.
        #
        # Please refer to issue: https://github.com/huggingface/transformers/issues/11565
        builder_kwargs = cls._get_kwargs(load_dataset_builder, op.hf_kwargs)
        builder = load_dataset_builder(op.path, **builder_kwargs)
        download_and_prepare_kwargs = cls._get_kwargs(
            DatasetBuilder.download_and_prepare, op.hf_kwargs
        )

        if op.single_data_file is not None:
            download_and_prepare_kwargs[
                "verification_mode"
            ] = VerificationMode.NO_CHECKS
            split = op.hf_kwargs["split"]
            split_data_files = builder.config.data_files[split]
            split_data_files[:] = [op.single_data_file]

        builder.download_and_prepare(**download_and_prepare_kwargs)
        as_dataset_kwargs = cls._get_kwargs(DatasetBuilder.as_dataset, op.hf_kwargs)
        ds = builder.as_dataset(**as_dataset_kwargs)
        ctx[op.outputs[0].key] = ds


def from_huggingface(
    path: str,
    name: Optional[str] = None,
    data_dir: Optional[str] = None,
    data_files: Optional[
        Union[str, Sequence[str], Mapping[str, Union[str, Sequence[str]]]]
    ] = None,
    split: Optional[Union[str, Split]] = None,
    cache_dir: Optional[str] = None,
    features: Optional[Features] = None,
    download_config: Optional[DownloadConfig] = None,
    download_mode: Optional[Union[DownloadMode, str]] = None,
    verification_mode: Optional[Union[VerificationMode, str]] = None,
    keep_in_memory: Optional[bool] = None,
    save_infos: bool = False,
    revision: Optional[Union[str, Version]] = None,
    token: Optional[Union[bool, str]] = None,
    streaming: bool = False,
    num_proc: Optional[int] = None,
    storage_options: Optional[Dict] = None,
    **config_kwargs,
) -> Dataset:
    """Create a dataset from a Hugging Face Datasets Dataset.

    This function is parallelized, and is intended to be used
    with Hugging Face Datasets that are loaded into memory (as opposed
    to memory-mapped).

    Args:
        path (`str`):
            Path or name of the dataset.
            Depending on `path`, the dataset builder that is used comes from a generic dataset script (JSON, CSV, Parquet, text etc.) or from the dataset script (a python file) inside the dataset directory.

            For local datasets:

            - if `path` is a local directory (containing data files only)
              -> load a generic dataset builder (csv, json, text etc.) based on the content of the directory
              e.g. `'./path/to/directory/with/my/csv/data'`.
            - if `path` is a local dataset script or a directory containing a local dataset script (if the script has the same name as the directory)
              -> load the dataset builder from the dataset script
              e.g. `'./dataset/squad'` or `'./dataset/squad/squad.py'`.

            For datasets on the Hugging Face Hub (list all available datasets with [`huggingface_hub.list_datasets`])

            - if `path` is a dataset repository on the HF hub (containing data files only)
              -> load a generic dataset builder (csv, text etc.) based on the content of the repository
              e.g. `'username/dataset_name'`, a dataset repository on the HF hub containing your data files.
            - if `path` is a dataset repository on the HF hub with a dataset script (if the script has the same name as the directory)
              -> load the dataset builder from the dataset script in the dataset repository
              e.g. `glue`, `squad`, `'username/dataset_name'`, a dataset repository on the HF hub containing a dataset script `'dataset_name.py'`.

        name (`str`, *optional*):
            Defining the name of the dataset configuration.
        data_dir (`str`, *optional*):
            Defining the `data_dir` of the dataset configuration. If specified for the generic builders (csv, text etc.) or the Hub datasets and `data_files` is `None`,
            the behavior is equal to passing `os.path.join(data_dir, **)` as `data_files` to reference all the files in a directory.
        data_files (`str` or `Sequence` or `Mapping`, *optional*):
            Path(s) to source data file(s).
        split (`Split` or `str`):
            Which split of the data to load.
            If `None`, will return a `dict` with all splits (typically `datasets.Split.TRAIN` and `datasets.Split.TEST`).
            If given, will return a single Dataset.
            Splits can be combined and specified like in tensorflow-datasets.
        cache_dir (`str`, *optional*):
            Directory to read/write data. Defaults to `"~/.cache/huggingface/datasets"`.
        features (`Features`, *optional*):
            Set the features type to use for this dataset.
        download_config ([`DownloadConfig`], *optional*):
            Specific download configuration parameters.
        download_mode ([`DownloadMode`] or `str`, defaults to `REUSE_DATASET_IF_EXISTS`):
            Download/generate mode.
        verification_mode ([`VerificationMode`] or `str`, defaults to `BASIC_CHECKS`):
            Verification mode determining the checks to run on the downloaded/processed dataset information (checksums/size/splits/...).

            .. versionadded:: 2.9.1
        keep_in_memory (`bool`, defaults to `None`):
            Whether to copy the dataset in-memory. If `None`, the dataset
            will not be copied in-memory unless explicitly enabled by setting `datasets.config.IN_MEMORY_MAX_SIZE` to
            nonzero. See more details in the [improve performance](../cache#improve-performance) section.
        save_infos (`bool`, defaults to `False`):
            Save the dataset information (checksums/size/splits/...).
        revision ([`Version`] or `str`, *optional*):
            Version of the dataset script to load.
            As datasets have their own git repository on the Datasets Hub, the default version "main" corresponds to their "main" branch.
            You can specify a different version than the default "main" by using a commit SHA or a git tag of the dataset repository.
        token (`str` or `bool`, *optional*):
            Optional string or boolean to use as Bearer token for remote files on the Datasets Hub.
            If `True`, or not specified, will get token from `"~/.huggingface"`.
        streaming (`bool`, defaults to `False`):
            If set to `True`, don't download the data files. Instead, it streams the data progressively while
            iterating on the dataset. An [`IterableDataset`] or [`IterableDatasetDict`] is returned instead in this case.

            Note that streaming works for datasets that use data formats that support being iterated over like txt, csv, jsonl for example.
            Json files may be downloaded completely. Also streaming from remote zip or gzip files is supported but other compressed formats
            like rar and xz are not yet supported. The tgz format doesn't allow streaming.
        num_proc (`int`, *optional*, defaults to `None`):
            Number of processes when downloading and generating the dataset locally.
            Multiprocessing is disabled by default.

            .. versionadded:: 2.7.0
        storage_options (`dict`, *optional*, defaults to `None`):
            **Experimental**. Key/value pairs to be passed on to the dataset file-system backend, if any.

            .. versionadded:: 2.11.0
        **config_kwargs (additional keyword arguments):
            Keyword arguments to be passed to the `BuilderConfig`
            and used in the [`DatasetBuilder`].

    Returns:
        Dataset
    """
    if split is None:
        raise Exception("Arg `split` is required for `from_huggingface`.")

    import datasets

    check_signature_compatible(
        from_huggingface,
        datasets.load_dataset,
        "Please use a compatible version of datasets.",
    )
    # For compatible different versions of API. e.g.
    # new API: from_huggingface(path, token=None)
    # old API: from_huggingface(path, use_auth_token=None)
    # then, from_huggingface("abc") can be forward from new API to old API
    # because non-compatible params are defaults values.
    kwargs = get_non_default_kwargs(locals(), from_huggingface)
    kwargs.pop("path")
    auto_rechunk = kwargs.pop("auto_rechunk", True)
    op = FromHuggingface(
        output_types=[OutputType.huggingface_dataset],
        path=path,
        auto_rechunk=auto_rechunk,
        hf_kwargs=kwargs,
    )
    return op().to_dataset()



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import functools
import json
import os.path
import threading
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import ExitStack
from queue import Queue
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from .._mars.utils import lazy_import

torch = lazy_import("torch")
if TYPE_CHECKING or torch is None:
    _TorchIterableDataset = object
else:
    _TorchIterableDataset = torch.utils.data.IterableDataset


@dataclasses.dataclass(init=False)
class GroupInfo:
    name: str
    path: str
    index: pa.Table
    schema: pa.Schema
    info: dict


@dataclasses.dataclass
class _FormatColumnInfo:
    as_py_columns: List[int] = dataclasses.field(default_factory=list)
    column_name_to_decoder: Dict[str, Callable] = dataclasses.field(
        default_factory=dict
    )


@dataclasses.dataclass
class _FinishSentinel:
    exception: Optional[Exception] = None


class Formatter:
    def format_batch(self, tables: List[pa.Table], indices: List[int]):
        result = {}
        for t in tables:
            d = t.take(indices).to_pydict()
            result.update(d)
        return result


class FeatureFormatter(Formatter):
    def __init__(self, features: Dict[str, Any], table_schemas: List[pa.Schema]):
        table_format_infos = []
        for ts in table_schemas:
            info = _FormatColumnInfo()
            for idx, name in enumerate(ts.names):
                if feat := features.get(name):
                    info.column_name_to_decoder[name] = feat.decode_example
                else:
                    info.as_py_columns.append(idx)
            table_format_infos.append(info)
        self._table_format_infos = table_format_infos

    def format_batch(self, tables: List[pa.Table], indices: List[int]):
        result = {}
        for info, table in zip(self._table_format_infos, tables):
            d = (
                table.select(info.as_py_columns).take(indices).to_pydict()
                if info.as_py_columns
                else {}
            )
            for name, decoder in info.column_name_to_decoder.items():
                chunked_array = table[name]
                d[name] = [
                    decoder(
                        {"bytes": chunked_array[idx]["bytes"].as_buffer(), "path": None}
                    )
                    for idx in indices
                ]
            result.update(d)
        return result


def _result_or_cancel(fut, timeout=None):
    try:
        try:
            return fut.result(timeout)
        finally:
            fut.cancel()
    finally:
        # Break a reference cycle with the exception in self._exception
        del fut


def map_retry(executor, fn, *iterables, timeout=None, retry=0):
    """Returns an iterator equivalent to map(fn, iter).

    Args:
        executor: A ThreadPoolExecutor instance.
        fn: A callable that will take as many arguments as there are
            passed iterables.
        timeout: The maximum number of seconds to wait. If None, then there
            is no limit on the wait time.
        retry: The retry times.

    Returns:
        An iterator equivalent to: map(func, *iterables) but the calls may
        be evaluated out-of-order.

    Raises:
        TimeoutError: If the entire result iterator could not be generated
            before the given timeout.
        Exception: If fn(*args) raises for any values.
    """
    if timeout is not None:
        end_time = timeout + time.monotonic()

    fs = []
    arg_list = []
    for args in zip(*iterables):
        arg_list.append(args)
        fs.append(executor.submit(fn, *args))

    # Yield must be hidden in closure so that the futures are submitted
    # before the first iterator value is required.
    def result_iterator():
        try:
            # reverse to keep finishing order
            fs.reverse()
            while fs:
                curr_args = arg_list.pop()
                # Careful not to keep a reference to the popped future
                for retry_time in range(retry + 1):
                    try:
                        if timeout is None:
                            yield _result_or_cancel(fs.pop())
                        else:
                            yield _result_or_cancel(
                                fs.pop(), end_time - time.monotonic()
                            )
                        break
                    except Exception as e:
                        if retry_time == retry:
                            raise e
                        fs.append(executor.submit(fn, *curr_args))
        finally:
            for future in fs:
                future.cancel()

    return result_iterator()


class IterableDataset(_TorchIterableDataset):
    _FORMAT_BATCH_SIZE = 10
    _DEFAULT_VERSION = "0.0.0"
    _META_DIR = ".meta"
    _FILE_NAME_FORMATTER = "{}_{}.arrow"

    def __init__(
        self,
        path: Union[str, os.PathLike],
        storage_options: Optional[dict] = None,
        prefetch: int = 2,
        fetch_retry: int = 2,
        fetch_timeout: float = 60,
        shuffle: bool = False,
        shuffle_seed: int = 9176,
        distributed_rank: Optional[int] = None,
        distributed_world_size: Optional[int] = None,
        worker_of_rank: Optional[int] = None,
        workers_per_rank: Optional[int] = None,
        num_threads: Optional[int] = None,
        version: Optional[str] = None,
    ):
        """
        An IterableDataset from the export path.

        Parameters
        ----------
        path: str
            The export path of Dataset include version.
        storage_options: dict
            Key/value pairs to be passed on to the caching file-system backend, if any.
        prefetch: int
            Prefetch chunks, default is 2.
        fetch_retry: int
            Fetch data retry times.
        fetch_timeout: float
            Fetch data timeout seconds.
        shuffle: bool
            Whether shuffle or not, default is False.
        shuffle_seed: int
            Shuffle seed, default is a fixed value.
        distributed_rank: int
            Distributed rank, if not set, use `RANK` from environment var.
        distributed_world_size: int
            Distributed world size, if not set, use `WORLD_SIZE` from environment var.
        worker_of_rank: int
            The worker id of rank, if not set, try to get from `torch.utils.data.get_worker_info`.
        workers_per_rank: int
            The num workers per rank, if not set, try to get from `torch.utils.data.get_worker_info`.
        num_threads: int
            The max worker threads for __iter__.
        version: str
            The version string, default is 0.0.0.
        """
        path = os.path.join(path, version or self._DEFAULT_VERSION)
        self._info, self._group_infos = self._get_infos(path, storage_options)
        self._path = path
        self._storage_options = storage_options
        self._prefetch = prefetch
        self._fetch_retry = fetch_retry
        self._fetch_timeout = fetch_timeout
        self._shuffle = shuffle
        self._shuffle_seed = shuffle_seed
        self._distributed_rank = distributed_rank
        self._distributed_world_size = distributed_world_size
        self._worker_of_rank = worker_of_rank
        self._workers_per_rank = workers_per_rank
        self._num_threads = num_threads
        self._epoch = 0

    def __iter__(self):
        import fsspec

        worker_index = self._get_worker_index()
        formatter = self._get_formatter()

        # Create a new fs instance in __iter__ make sure the fs instance
        # is bound to worker.
        fs_token_paths = fsspec.get_fs_token_paths(
            self._path, storage_options=self._storage_options
        )
        fs: fsspec.AbstractFileSystem = fs_token_paths[0]
        # The flag to stop the _formatter thread.
        finish = False
        # A sentinel to mark the upstream is finish.
        finish_sentinel = _FinishSentinel()
        # The queue _formatter -> __iter__
        format_queue = Queue(
            maxsize=self._info["max_chunk_rows"] // self._FORMAT_BATCH_SIZE + 1
        )

        def _load_arrow_table(filepath):
            # TODO(codingl2k1): mmap if local.
            with fs.open(filepath, "rb") as f:
                with pa.ipc.RecordBatchStreamReader(f) as reader:
                    return reader.read_all()

        def _prefetcher(_executor: ThreadPoolExecutor):
            for idx in worker_index:
                group_files = [
                    os.path.join(
                        group.path,
                        self._FILE_NAME_FORMATTER.format(*group.index[0][idx].as_py()),
                    )
                    for group in self._group_infos
                ]
                yield map_retry(
                    _executor,
                    _load_arrow_table,
                    group_files,
                    timeout=self._fetch_timeout,
                    retry=self._fetch_retry,
                )

        def _formatter(_output_queue: Queue):
            # TODO(codingl2k1): Use as_completed to optimize shuffle True.
            rng = (
                np.random.default_rng(self._shuffle_seed + self._epoch)
                if self._shuffle
                else None
            )
            with ThreadPoolExecutor(
                max_workers=self._num_threads,
                thread_name_prefix=self.__iter__.__qualname__,
            ) as executor:
                try:
                    prefetch = _prefetcher(executor)
                    buffer = [next(prefetch, None) for _ in range(self._prefetch)]
                    idx = 0
                    buffer_len = len(buffer)
                    while not finish:
                        tables = buffer[idx]
                        if tables is None:
                            break
                        buffer[idx] = next(prefetch, None)
                        idx += 1
                        idx %= buffer_len

                        tables = list(tables)
                        num_chunk_rows = len(tables[0])
                        in_chunk_index = np.arange(num_chunk_rows)
                        if rng is not None:
                            rng.shuffle(in_chunk_index)
                        for indices in np.array_split(
                            in_chunk_index,
                            num_chunk_rows // self._FORMAT_BATCH_SIZE + 1,
                        ):
                            _output_queue.put(
                                executor.submit(formatter.format_batch, tables, indices)
                            )
                except Exception as ex:
                    finish_sentinel.exception = ex
                finally:
                    _output_queue.put(finish_sentinel)

        format_thread = threading.Thread(target=_formatter, args=(format_queue,))
        format_thread.start()
        try:
            while fut := format_queue.get():
                if fut is finish_sentinel:
                    break
                for example in self._batch_to_examples(fut.result()):
                    yield example
        except Exception as e:
            finish = True
            finish_sentinel.exception = e
        finally:
            format_thread.join()
        # Reraise the exception.
        if finish_sentinel.exception is not None:
            raise finish_sentinel.exception

    def __len__(self):
        """Get data length according to rank, world size, worker id and num workers."""
        if self._get_world_size() > 1:
            worker_index = self._get_worker_index()
            group = self._group_infos[0]
            num_rows_series = group.index[f"{group.name}.num_rows"]
            return pc.sum(num_rows_series.take(worker_index)).as_py()
        else:
            return self._info["num_rows"]

    @staticmethod
    def _batch_to_examples(
        batch: Dict[str, list]
    ) -> Generator[Dict[str, Any], None, None]:
        """Convert a batch (dict of examples) to examples list"""
        n_examples = len(batch[next(iter(batch))])
        for i in range(n_examples):
            yield {col: array[i] for col, array in batch.items()}

    @classmethod
    def _get_infos(cls, path, storage_options) -> Tuple[Dict, List[GroupInfo]]:
        import fsspec

        # TODO(codingl2k1): Merge group meta files into one.
        group_infos = []
        futures = []
        with ThreadPoolExecutor(
            thread_name_prefix=IterableDataset._get_infos.__qualname__
        ) as executor, ExitStack() as es:
            fs_token_paths = fsspec.get_fs_token_paths(
                path, storage_options=storage_options
            )
            fs: fsspec.AbstractFileSystem = fs_token_paths[0]
            path = fs_token_paths[2][0]
            es.callback(fs.clear_instance_cache)

            with fs.open(os.path.join(path, "info.json"), "r") as f:
                info = json.load(f)

            for name in info["groups"]:
                group_path = os.path.join(path, name)
                group_meta_path = os.path.join(group_path, cls._META_DIR)
                group_info = GroupInfo()
                group_info.name = name
                group_info.path = group_path
                group_infos.append(group_info)

                def _fill_index(_path, _group_info):
                    with fs.open(_path, "rb") as f:
                        with pa.ipc.RecordBatchStreamReader(f) as reader:
                            _group_info.index = reader.read_all()

                fut = executor.submit(
                    _fill_index,
                    os.path.join(group_meta_path, "index.arrow"),
                    group_info,
                )
                futures.append(fut)

                def _fill_schema(_path, _group_info):
                    with fs.open(_path, "rb") as f:
                        with pa.ipc.RecordBatchStreamReader(f) as reader:
                            _group_info.schema = reader.schema

                fut = executor.submit(
                    _fill_schema,
                    os.path.join(group_meta_path, "schema.arrow"),
                    group_info,
                )
                futures.append(fut)

                def _fill_info(_path, _group_info):
                    with fs.open(_path, "r") as f:
                        _group_info.info = json.load(f)

                fut = executor.submit(
                    _fill_info, os.path.join(group_meta_path, "info.json"), group_info
                )
                futures.append(fut)
            # Raise exception if exists.
            for fut in as_completed(futures):
                fut.result()
        return info, group_infos

    def _get_rank(self) -> int:
        """Returns the rank of the current process, which is on ``[0; WORLD_SIZE - 1]``.

        Returns:
            int: The rank.
        """
        rank = self._distributed_rank
        if rank is None:
            return int(os.environ.get("RANK", 0))
        return rank

    def _get_world_size(self) -> int:
        """Returns the world size, which is the number of processes participating in this training run.

        Returns:
            int: The world size.
        """
        world_size = self._distributed_world_size
        if world_size is None:
            return int(os.environ.get("WORLD_SIZE", 1))
        return world_size

    def _get_worker_info(self):
        """Get worker id and num workers."""
        worker_id = self._worker_of_rank
        num_workers = self._workers_per_rank
        if worker_id is None or num_workers is None:
            worker_info = torch and torch.utils.data.get_worker_info()
            if worker_info is None:
                worker_id = 0
                num_workers = 1
            else:
                worker_id = worker_info.id
                num_workers = worker_info.num_workers
        return worker_id, num_workers

    def _get_worker_index(self):
        rank, world_size = self._get_rank(), self._get_world_size()
        assert rank < world_size

        # Generate an index for the Dataset group index.
        num_group_chunks = len(self._group_infos[0].index)
        rank_index = np.arange(num_group_chunks)
        if self._shuffle:
            rng = np.random.default_rng(self._shuffle_seed + self._epoch)
            rng.shuffle(rank_index)
        # Get each worker index by rank and world size.
        rank_index = rank_index[rank:num_group_chunks:world_size]
        worker_id, num_workers = self._get_worker_info()
        return (
            rank_index[worker_id : len(rank_index) : num_workers]
            if num_workers > 1
            else rank_index
        )

    def _get_formatter(self) -> Formatter:
        """
        Get formatters from metadata.

        Returns
        -------
            The Formatter instance.
        """
        from datasets.features.audio import Audio
        from datasets.features.features import generate_from_dict
        from datasets.features.image import Image

        # TODO(codingl2k1): Impl our decoders.
        metadata = self.schema.metadata.get(b"huggingface")
        if metadata:
            metadata = json.loads(metadata.decode())
            metadata = generate_from_dict(metadata)
            features = metadata["info"]["features"]
            requires_decoding = {
                k: v for k, v in features.items() if isinstance(v, (Audio, Image))
            }
            if requires_decoding:
                return FeatureFormatter(
                    requires_decoding,
                    [group.schema for group in self._group_infos],
                )
        # Fast path without decoding.
        return Formatter()

    @functools.cached_property
    def groups(self) -> List[str]:
        return self._info["groups"]

    @functools.cached_property
    def schema(self) -> pa.Schema:
        return pa.unify_schemas([gi.schema for gi in self._group_infos])

    @functools.cached_property
    def shape(self) -> Tuple:
        return len(self), len(self.schema.names)

    def info(self) -> Dict:
        return self._info

    def group_infos(self) -> List[GroupInfo]:
        """Groups has order, so returns a list."""
        return self._group_infos

    def set_epoch(self, epoch: int):
        self._epoch = epoch

    def epoch(self) -> int:
        return self._epoch



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from typing import Any, Union

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...core.context import Context
from ...core.custom_log import redirect_custom_log
from ...core.operand import MapReduceOperand, OperandStage
from ...serialization.serializables import (
    AnyField,
    BoolField,
    Int32Field,
    ListField,
    StringField,
)
from ...utils import enter_current_session
from ..operands import DataFrameOperandMixin, DataFrameShuffleProxy
from ..utils import build_concatenated_rows_frame, hash_dataframe_on, parse_index


class DataFramePivotTable(MapReduceOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.PIVOT_TABLE

    values = AnyField("values", default=None)
    index = AnyField("index", default=None)
    columns = AnyField("columns", default=None)
    aggfunc = AnyField("aggfunc", default="mean")
    fill_value = AnyField("fill_value", default=None)
    margins = BoolField("margins", default=False)
    dropna = BoolField("drapna", default=True)
    margins_name = StringField("margins_name", default="All")
    observed = BoolField("observed", default=False)
    sort = BoolField("sort", default=True)
    shuffle_size = Int32Field("shuffle_size")
    output_columns = ListField("output_columns", default=None)

    @classmethod
    def execute_map(cls, ctx: Union[dict, Context], op: "DataFramePivotTable"):
        chunk = op.outputs[0]
        df = ctx[op.inputs[0].key]
        filters = hash_dataframe_on(df, op.index, op.shuffle_size)

        for index_idx, index_filter in enumerate(filters):
            reducer_index = (index_idx, chunk.index[1])
            ctx[chunk.key, reducer_index] = (
                ctx.get_current_chunk().index,
                df.iloc[index_filter],
            )

    @classmethod
    def execute_reduce(cls, ctx: Union[dict, Context], op: "DataFramePivotTable"):
        chunk = op.outputs[0]
        input_idx_to_df = dict(op.iter_mapper_data(ctx))
        row_idxes = sorted(input_idx_to_df.keys())

        res = []
        for row_idx in row_idxes:
            row_df = input_idx_to_df.get(row_idx, None)
            if row_df is not None:
                res.append(row_df)

        ctx[chunk.key] = pd.concat(res, axis=0)

    @classmethod
    def execute_combine(cls, ctx: Union[dict, Context], op: "DataFramePivotTable"):
        input_data = ctx[op.inputs[0].key]
        out = op.outputs[0]

        for dtype in op.output_columns:
            if dtype not in input_data.dtypes.index:
                input_data[dtype] = np.nan if op.fill_value is None else op.fill_value

        ctx[out.key] = input_data

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx: Union[dict, Context], op: "DataFramePivotTable"):
        if op.stage == OperandStage.map:
            cls.execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            cls.execute_reduce(ctx, op)
        elif op.stage == OperandStage.combine:
            cls.execute_combine(ctx, op)
        else:
            input_data = ctx[op.inputs[0].key]
            out = op.outputs[0]
            result = input_data.pivot_table(
                values=op.values,
                index=op.index,
                columns=op.columns,
                aggfunc=op.aggfunc,
                fill_value=op.fill_value,
                margins=op.margins,
                dropna=op.dropna,
                observed=op.observed,
                sort=op.sort,
            )
            ctx[out.key] = result

    @classmethod
    def tile_one_chunk(cls, op: "DataFramePivotTable"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_df.chunks:
            new_op = op.copy().reset_key()
            new_op.tileable_op_key = op.key
            chunks.append(
                new_op.new_chunk(
                    [c],
                    shape=(np.nan, np.nan),
                    index=c.index,
                    dtypes=out_df.dtypes,
                    index_value=c.index_value,
                    columns_value=c.columns_value,
                )
            )

        new_op = op.copy()
        kw = out_df.params.copy()
        new_nsplits = ((np.nan,), (np.nan,))
        kw.update(dict(chunks=chunks, nsplits=new_nsplits))
        return new_op.new_tileables(op.inputs, **kw)

    @classmethod
    def tile(cls, op: "DataFramePivotTable"):
        in_df = build_concatenated_rows_frame(op.inputs[0])

        if len(in_df.chunks) == 1:
            return cls.tile_one_chunk(op)

        output_type = OutputType.dataframe
        chunk_shape = in_df.chunk_shape

        # generate map chunks
        map_chunks = []
        for chunk in in_df.chunks:
            map_op = op.copy().reset_key()
            map_op.stage = OperandStage.map
            map_op.shuffle_size = chunk_shape[0]
            map_op._output_types = [output_type]
            chunk_inputs = [chunk]
            map_chunks.append(
                map_op.new_chunk(
                    chunk_inputs,
                    shape=(np.nan, np.nan),
                    index=chunk.index,
                )
            )
        proxy_chunk = DataFrameShuffleProxy(output_types=[output_type]).new_chunk(
            map_chunks, shape=()
        )

        # generate reduce chunks
        reduce_chunks = []
        out_indices = list(itertools.product(*(range(s) for s in chunk_shape)))
        for ordinal, out_idx in enumerate(out_indices):
            reduce_op = op.copy().reset_key()
            reduce_op._output_types = [output_type]
            reduce_op.stage = OperandStage.reduce
            reduce_op.reducer_ordinal = ordinal
            reduce_op.n_reducers = len(out_indices)
            reduce_chunks.append(
                reduce_op.new_chunk(
                    [proxy_chunk], shape=(np.nan, np.nan), index=out_idx
                )
            )

        # generate output chunks
        out_chunks = []
        for chunk in reduce_chunks:
            new_op = op.copy().reset_key()
            new_shape = (np.nan, np.nan)
            params = dict(shape=new_shape, index=chunk.index)
            params.update(
                dict(
                    dtypes=in_df.dtypes,
                    columns_value=in_df.columns_value,
                    index_value=parse_index(None, chunk.key, proxy_chunk.key),
                )
            )
            out_chunks.append(new_op.new_chunk([chunk], **params))
        yield out_chunks

        filtered_chunks = [chunk for chunk in out_chunks if chunk.shape[1] != 0]
        for i, chunk in enumerate(filtered_chunks):
            chunk._index = (i, 0)

        combined_dtypes = pd.concat(
            [chunk.dtypes for chunk in filtered_chunks]
        ).to_dict()
        output_columns = list(combined_dtypes.keys())
        output_dtypes = pd.Series(
            combined_dtypes.values(),
            index=pd.Index(
                combined_dtypes.keys(), name=filtered_chunks[0].dtypes.index.name
            ),
        )
        out_columns_value = parse_index(output_dtypes.index, store_data=True)

        # generate combine chunks
        combine_chunks = []
        for chunk in filtered_chunks:
            if chunk.shape[1] < len(output_columns):
                combine_op = op.copy().reset_key()
                combine_op.stage = OperandStage.combine
                combine_op.output_columns = output_columns
                params = dict(
                    shape=(chunk.shape[0], len(output_columns)),
                    index=chunk.index,
                    dtypes=output_dtypes,
                    columns_value=out_columns_value,
                    index_value=chunk.index_value,
                )
                combine_chunks.append(combine_op.new_chunk([chunk], **params))
            else:
                combine_chunks.append(chunk)

        new_op = op.copy()
        kw = dict(
            columns_value=out_columns_value,
            dtypes=output_dtypes,
            index_value=filtered_chunks[0].index_value,
        )
        new_nsplits = (
            tuple(chunk.shape[0] for chunk in combine_chunks),
            (len(output_columns),),
        )
        kw.update(dict(chunks=combine_chunks, nsplits=new_nsplits))

        return new_op.new_tileables(op.inputs, **kw)

    def __call__(self, df):
        return self.new_dataframe([df])


def df_pivot_table(
    df: pd.DataFrame,
    values: Any = None,
    index: Any = None,
    columns: Any = None,
    aggfunc="mean",
    fill_value=None,
    margins: bool = False,
    dropna: bool = True,
    margins_name: str = "All",
    observed: bool = False,
    sort: bool = True,
) -> pd.DataFrame:
    """
    Create a spreadsheet-style pivot table as a DataFrame.

    The levels in the pivot table will be stored in MultiIndex objects
    (hierarchical indexes) on the index and columns of the result DataFrame.

    Parameters
    ----------
    values : list-like or scalar, optional
        Column or columns to aggregate.
    index : column, Grouper, array, or list of the previous
        If an array is passed, it must be the same length as the data. The
        list can contain any of the other types (except list).
        Keys to group by on the pivot table index.  If an array is passed,
        it is being used as the same manner as column values.
    columns : column, Grouper, array, or list of the previous
        If an array is passed, it must be the same length as the data. The
        list can contain any of the other types (except list).
        Keys to group by on the pivot table column.  If an array is passed,
        it is being used as the same manner as column values.
    aggfunc : function, list of functions, dict, default numpy.mean
        If list of functions passed, the resulting pivot table will have
        hierarchical columns whose top level are the function names
        (inferred from the function objects themselves)
        If dict is passed, the key is column to aggregate and value
        is function or list of functions. If ``margin=True``,
        aggfunc will be used to calculate the partial aggregates.
    fill_value : scalar, default None
        Value to replace missing values with (in the resulting pivot table,
        after aggregation).
    margins : bool, default False
        If ``margins=True``, special ``All`` columns and rows
        will be added with partial group aggregates across the categories
        on the rows and columns. This feature is currently not supported
        in xorbits.
    dropna : bool, default True
        Do not include columns whose entries are all NaN. If True,
        rows with a NaN value in any column will be omitted before
        computing margins. This feature is currently not supported in
        xorbits.
    margins_name : str, default 'All'
        Name of the row / column that will contain the totals
        when margins is True. This feature is currently not supported in
        xorbits.
    observed : bool, default False
        This only applies if any of the groupers are Categoricals.
        If True: only show observed values for categorical groupers.
        If False: show all values for categorical groupers. This
        feature is currently not supported in xorbits.

    sort : bool, default True
        Specifies if the result should be sorted.

        .. versionadded:: 1.3.0

    Returns
    -------
    DataFrame
        An Excel style pivot table.

    See Also
    --------
    DataFrame.pivot : Pivot without aggregation that can handle
        non-numeric data.
    DataFrame.melt: Unpivot a DataFrame from wide to long format,
        optionally leaving identifiers set.
    wide_to_long : Wide panel to long format. Less flexible but more
        user-friendly than melt.

    Notes
    -----
    Reference :ref:`the user guide <reshaping.pivot>` for more examples.

    Examples
    --------
    >>> df = pd.DataFrame({"A": ["foo", "foo", "foo", "foo", "foo",
    ...                          "bar", "bar", "bar", "bar"],
    ...                    "B": ["one", "one", "one", "two", "two",
    ...                          "one", "one", "two", "two"],
    ...                    "C": ["small", "large", "large", "small",
    ...                          "small", "large", "small", "small",
    ...                          "large"],
    ...                    "D": [1, 2, 2, 3, 3, 4, 5, 6, 7],
    ...                    "E": [2, 4, 5, 5, 6, 6, 8, 9, 9]})
    >>> df
            A    B      C  D  E
    0  foo  one  small  1  2
    1  foo  one  large  2  4
    2  foo  one  large  2  5
    3  foo  two  small  3  5
    4  foo  two  small  3  6
    5  bar  one  large  4  6
    6  bar  one  small  5  8
    7  bar  two  small  6  9
    8  bar  two  large  7  9

    This first example aggregates values by taking the sum.

    >>> table = pd.pivot_table(df, values='D', index=['A', 'B'],
    ...                        columns=['C'], aggfunc=np.sum)
    >>> table
    C        large  small
    A   B
    bar one    4.0    5.0
        two    7.0    6.0
    foo one    4.0    1.0
        two    NaN    6.0

    We can also fill missing values using the `fill_value` parameter.

    >>> table = pd.pivot_table(df, values='D', index=['A', 'B'],
    ...                        columns=['C'], aggfunc=np.sum, fill_value=0)
    >>> table
    C        large  small
    A   B
    bar one      4      5
        two      7      6
    foo one      4      1
        two      0      6

    The next example aggregates by taking the mean across multiple columns.

    >>> table = pd.pivot_table(df, values=['D', 'E'], index=['A', 'C'],
    ...                        aggfunc={'D': np.mean, 'E': np.mean})
    >>> table
                    D         E
    A   C
    bar large  5.500000  7.500000
        small  5.500000  8.500000
    foo large  2.000000  4.500000
        small  2.333333  4.333333

    We can also calculate multiple types of aggregations for any given
    value column.

    >>> table = pd.pivot_table(df, values=['D', 'E'], index=['A', 'C'],
    ...                        aggfunc={'D': np.mean,
    ...                                 'E': [min, max, np.mean]})
    >>> table
                        D   E
                    mean max      mean  min
    A   C
    bar large  5.500000   9  7.500000    6
        small  5.500000   9  8.500000    8
    foo large  2.000000   5  4.500000    4
        small  2.333333   6  4.333333    2
    """
    if index is not None and not isinstance(index, (str, list, pd.Index)):
        raise NotImplementedError(
            "The 'index' parameter should be of type str or list of str."
        )

    if columns is not None and not isinstance(columns, (str, list, pd.Index)):
        raise NotImplementedError(
            "The 'columns' parameter should be of type str or list of str."
        )

    if margins:
        raise NotImplementedError(
            "The 'margins=True' configuration is not currently supported in this version of xorbits."
        )

    if observed:
        raise NotImplementedError(
            "The 'observed=True' configuration is not currently supported in this version of xorbits."
        )

    if dropna == False:
        raise NotImplementedError(
            "The 'dropna=False' configuration is not currently supported in this version of xorbits."
        )

    op = DataFramePivotTable(
        values=values,
        index=index,
        columns=columns,
        aggfunc=aggfunc,
        fill_value=fill_value,
        margins=margins,
        dropna=dropna,
        margins_name=margins_name,
        observed=observed,
        sort=sort,
    )
    return op(df)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

from ... import opcodes as OperandDef
from ...utils import classproperty
from .core import DataFrameBinopUfunc
from .docstring import bin_arithmetic_doc


class DataFrameSubtract(DataFrameBinopUfunc):
    _op_type_ = OperandDef.SUB

    _func_name = "sub"
    _rfunc_name = "rsub"

    @classproperty
    def _operator(self):
        return operator.sub

    @classproperty
    def tensor_op_type(self):
        from ...tensor.arithmetic import TensorSubtract

        return TensorSubtract


_sub_example = """
>>> a.subtract(b, fill_value=0).execute()
a    0.0
b    1.0
c    1.0
d   -1.0
e    NaN
dtype: float64
"""


@bin_arithmetic_doc("Subtraction", equiv="-", series_example=_sub_example)
def subtract(df, other, axis="columns", level=None, fill_value=None):
    op = DataFrameSubtract(
        axis=axis, level=level, fill_value=fill_value, lhs=df, rhs=other
    )
    return op(df, other)


@bin_arithmetic_doc("Subtraction", equiv="-", series_example=_sub_example)
def rsubtract(df, other, axis="columns", level=None, fill_value=None):
    op = DataFrameSubtract(
        axis=axis, level=level, fill_value=fill_value, lhs=other, rhs=df
    )
    return op.rcall(df, other)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC, abstractmethod
from typing import Dict, List

from ..storage.base import StorageLevel
from ..typing import BandType, SessionType
from ..utils import classproperty


class Context(ABC):
    """
    Context that providing API that can be
    used inside `tile` and `execute`.
    """

    all_contexts = []

    def __init__(
        self,
        session_id: str = None,
        supervisor_address: str = None,
        worker_address: str = None,
        local_address: str = None,
        band: BandType = None,
    ):
        if session_id is None:
            # try to get session id from environment
            session_id = os.environ.get("MARS_SESSION_ID")
            if session_id is None:
                raise ValueError("session_id should be provided to create a context")
        if supervisor_address is None:
            # try to get supervisor address from environment
            supervisor_address = os.environ.get("MARS_SUPERVISOR_ADDRESS")
            if supervisor_address is None:
                raise ValueError(
                    "supervisor_address should be provided to create a context"
                )

        self.session_id = session_id
        self.supervisor_address = supervisor_address
        self.worker_address = worker_address
        self.local_address = local_address
        self.band = band

    @abstractmethod
    def get_current_session(self) -> SessionType:
        """
        Get current session

        Returns
        -------
        session
        """

    @abstractmethod
    def get_local_host_ip(self) -> str:
        """
        Get local worker's host ip

        Returns
        -------
        host_ip : str
        """

    @abstractmethod
    def get_supervisor_addresses(self) -> List[str]:
        """
        Get supervisor addresses.

        Returns
        -------
        supervisor_addresses : list
        """

    @abstractmethod
    def get_worker_addresses(self) -> List[str]:
        """
        Get worker addresses.

        Returns
        -------
        worker_addresses : list
        """

    @abstractmethod
    def get_worker_bands(self) -> List[BandType]:
        """
        Get worker bands.

        Returns
        -------
        worker_bands : list
        """

    @abstractmethod
    def get_total_n_cpu(self) -> int:
        """
        Get number of cpus.

        Returns
        -------
        number_of_cpu: int
        """

    @abstractmethod
    def get_slots(self) -> int:
        """
        Get num of slots of current band

        Returns
        -------
        number_of_bands: int
        """

    @abstractmethod
    def get_chunks_result(self, data_keys: List[str], fetch_only: bool = False) -> List:
        """
        Get result of chunks.

        Parameters
        ----------
        data_keys : list
            Data keys.
        fetch_only : bool
            If fetch_only, only fetch data but not return.

        Returns
        -------
        results : list
            Result of chunks if not fetch_only, else return None
        """

    @abstractmethod
    def get_chunks_meta(
        self, data_keys: List[str], fields: List[str] = None, error="raise"
    ) -> List[Dict]:
        """
        Get meta of chunks.

        Parameters
        ----------
        data_keys : list
            Data keys.
        fields : list
            Fields to filter.
        error : str
            raise, ignore

        Returns
        -------
        meta_list : list
            Meta list.
        """

    @abstractmethod
    def get_storage_info(self, address: str, level: StorageLevel):
        """
        Get the customized storage backend info of requested storage backend level at given worker.

        Parameters
        ----------
        address: str
            The worker address.
        level: StorageLevel
            The storage level to fetch the backend info.

        Returns
        -------
        info: dict
            Customized storage backend info dict of all workers. The key is
            worker address, the value is the backend info dict.
        """

    @abstractmethod
    def create_remote_object(self, name: str, object_cls, *args, **kwargs):
        """
        Create remote object.

        Parameters
        ----------
        name : str
            Object name.
        object_cls
            Object class.
        args
        kwargs

        Returns
        -------
        ref
        """

    @abstractmethod
    def get_remote_object(self, name: str):
        """
        Get remote object

        Parameters
        ----------
        name : str
            Object name.

        Returns
        -------
        ref
        """

    @abstractmethod
    def destroy_remote_object(self, name: str):
        """
        Destroy remote object.

        Parameters
        ----------
        name : str
            Object name.
        """

    @abstractmethod
    def register_custom_log_path(
        self,
        session_id: str,
        tileable_op_key: str,
        chunk_op_key: str,
        worker_address: str,
        log_path: str,
    ):
        """
        Register custom log path.

        Parameters
        ----------
        session_id : str
            Session ID.
        tileable_op_key : str
            Key of tileable's op.
        chunk_op_key : str
            Kye of chunk's op.
        worker_address : str
            Worker address.
        log_path : str
            Log path.
        """

    def new_custom_log_dir(self) -> str:
        """
        New custom log dir.

        Returns
        -------
        custom_log_dir : str
            Custom log dir.
        """

    def set_running_operand_key(self, session_id: str, op_key: str):
        """
        Set key of running operand.

        Parameters
        ----------
        session_id : str
        op_key : str
        """

    def set_progress(self, progress: float):
        """
        Set progress of running operand.

        Parameters
        ----------
        progress : float
        """

    def __enter__(self):
        Context.all_contexts.append(self)

    def __exit__(self, *_):
        Context.all_contexts.pop()

    @classproperty
    def current(cls):
        return cls.all_contexts[-1] if cls.all_contexts else None


def set_context(context: Context):
    Context.all_contexts.append(context)


def get_context() -> Context:
    return Context.current



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.model_selection as sk_ml

from ...._mars.learn import model_selection as mars_ml
from ...._mars.learn.model_selection import KFold as MarsKFold
from ...._mars.learn.model_selection import ParameterGrid as MarsParameterGrid
from ....core.utils.docstring import attach_module_callable_docstring
from ...utils import SKLearnBase, _collect_module_callables, _install_cls_members


class KFold(SKLearnBase):
    _marscls = MarsKFold


class ParameterGrid(SKLearnBase):
    _marscls = MarsParameterGrid

    def __len__(self):
        return len(self.mars_instance)

    def __iter__(self):
        return iter(self.mars_instance)

    def __getitem__(self, index):
        return self.mars_instance[index]


SKLEARN_ML_CLS_MAP = {
    KFold: MarsKFold,
    ParameterGrid: MarsParameterGrid,
}

MARS_SKLEARN_ML_CALLABLES = _collect_module_callables(
    mars_ml, sk_ml, skip_members=["register_op"]
)
_install_cls_members(SKLEARN_ML_CLS_MAP, MARS_SKLEARN_ML_CALLABLES, sk_ml)
attach_module_callable_docstring(KFold, sk_ml, sk_ml.KFold)
attach_module_callable_docstring(ParameterGrid, sk_ml, sk_ml.ParameterGrid)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Callable

from ...core.adapter import from_mars, mars_remote, to_mars


def _wrap_remote_func(c: Callable):
    """
    Making the inputs and outputs of the wrapped callable be xorbits types.
    """

    @functools.wraps(c)
    def wrapped(*args, **kwargs):
        return to_mars(c(*from_mars(args), **from_mars(kwargs)))

    return wrapped


def spawn(
    func,
    args=(),
    kwargs=None,
    retry_when_fail=False,
    n_output=None,
    output_types=None,
    **kw,
):
    """
    Spawn a function and return a Mars Object which can be executed later.

    Parameters
    ----------
    func : function
        Function to spawn.
    args: tuple
       Args to pass to function
    kwargs: dict
       Kwargs to pass to function
    retry_when_fail: bool, default False
       If True, retry when function failed.
    n_output: int
       Count of outputs for the function
    output_types: str or list, default "object"
        Specify type of returned objects.

    Returns
    -------
    Object
        Mars Object.

    Examples
    --------
    >>> import xorbits.remote as xr
    >>> def inc(x):
    ...     return x + 1
    >>>
    >>> result = xr.spawn(inc, args=(0,))
    >>> result.to_object()
    1


    >>> results = [xr.spawn(inc, args=(i,)) for i in range(10)]
    >>> [r.to_object() for r in results]
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Xorbits objects returned by :meth:`xorbits.remote.spawn` can be used as arguments when spawning
     another function.

    >>> def sum_all(xs):
    ...     return sum(xs)
    >>> xr.spawn(sum_all, args=(results,)).to_object()
    55

    functions can be spawned recursively.

    >>> def driver():
    ...     results = [xr.spawn(inc, args=(i,)) for i in range(10)]
    ...     return [r.to_object() for r in results]
    >>> xr.spawn(driver).to_object()
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Xorbits objects can also be used in spawned functions.

    >>> import xorbits.numpy as np
    >>> def driver2():
    ...     t = np.ones((10, 10))
    ...     return t.sum().to_numpy()
    >>> xr.spawn(driver2).to_object()
    100.0

    The argument `n_output` indicates that the function to spawn has multiple outputs.

    >>> def triage(alist):
    ...     ret = [], []
    ...     for i in alist:
    ...         if i < 0.5:
    ...             ret[0].append(i)
    ...         else:
    ...             ret[1].append(i)
    ...     return ret
    >>>
    >>> def sum_all(xs):
    ...     return sum(xs)
    >>>
    >>> l = [0.4, 0.7, 0.2, 0.8]
    >>> la, lb = xr.spawn(triage, args=(l,), n_output=2)
    >>>
    >>> sa = xr.spawn(sum_all, args=(la,))
    >>> sb = xr.spawn(sum_all, args=(lb,))
    >>> [sa.to_object(), sb.to_object()]
    [0.6000000000000001, 1.5]
    """

    func = _wrap_remote_func(func)
    args = to_mars(args)
    kwargs = to_mars(kwargs)

    ret = mars_remote.spawn(
        func,
        args,
        kwargs,
        retry_when_fail=retry_when_fail,
        n_output=n_output,
        output_types=output_types,
        **kw,
    )

    return from_mars(ret)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Union

import xoscar as mo

from ....core import Tileable
from ....lib.aio import alru_cache
from ...subtask import SubtaskResult
from ..core import MapReduceInfo, TaskResult, TileableGraph
from ..supervisor.manager import TaskManagerActor
from ..task_info_collector import TaskInfoCollectorActor
from .core import AbstractTaskAPI


class TaskAPI(AbstractTaskAPI):
    def __init__(
        self,
        session_id: str,
        task_info_collector_ref: mo.ActorRefType[TaskInfoCollectorActor],
        task_manager_ref: mo.ActorRefType[TaskManagerActor],
    ):
        self._session_id = session_id
        self._task_info_collector_ref = task_info_collector_ref
        self._task_manager_ref = task_manager_ref

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(
        cls, session_id: str, supervisor_address: str = None, local_address: str = None
    ) -> "TaskAPI":
        """
        Create Task API.

        Parameters
        ----------
        session_id : str
            Session ID.
        supervisor_address : str
            Supervisor address.
        local_address : str
            Local address.

        Returns
        -------
        task_api
            Task API.
        """
        try:
            task_manager_ref = await mo.actor_ref(
                supervisor_address, TaskManagerActor.gen_uid(session_id)
            )
        except (mo.ActorNotExist, ValueError):
            task_manager_ref = None
        try:
            task_info_collector_ref = await mo.actor_ref(
                local_address, TaskInfoCollectorActor.default_uid()
            )
        except (mo.ActorNotExist, ValueError):
            task_info_collector_ref = None

        return TaskAPI(session_id, task_info_collector_ref, task_manager_ref)

    async def get_task_results(self, progress: bool = False) -> List[TaskResult]:
        return await self._task_manager_ref.get_task_results(progress)

    async def submit_tileable_graph(
        self,
        graph: TileableGraph,
        fuse_enabled: bool = None,
        extra_config: dict = None,
    ) -> str:
        try:
            return await self._task_manager_ref.submit_tileable_graph(
                graph, fuse_enabled=fuse_enabled, extra_config=extra_config
            )
        except mo.ActorNotExist:
            raise RuntimeError("Session closed already")

    async def get_tileable_graph_as_json(self, task_id: str):
        return await self._task_manager_ref.get_tileable_graph_dict_by_task_id(task_id)

    async def get_tileable_details(self, task_id: str):
        return await self._task_manager_ref.get_tileable_details(task_id)

    async def get_tileable_subtasks(
        self, task_id: str, tileable_id: str, with_input_output: bool
    ):
        return await self._task_manager_ref.get_tileable_subtasks(
            task_id, tileable_id, with_input_output
        )

    async def wait_task(self, task_id: str, timeout: float = None):
        return await self._task_manager_ref.wait_task(task_id, timeout=timeout)

    async def get_task_result(self, task_id: str) -> TaskResult:
        return await self._task_manager_ref.get_task_result(task_id)

    async def get_task_progress(self, task_id: str) -> float:
        return await self._task_manager_ref.get_task_progress(task_id)

    async def cancel_task(self, task_id: str):
        return await self._task_manager_ref.cancel_task(task_id)

    async def get_fetch_tileables(self, task_id: str) -> List[Tileable]:
        return await self._task_manager_ref.get_task_result_tileables(task_id)

    async def set_subtask_result(self, subtask_result: SubtaskResult):
        return await self._task_manager_ref.set_subtask_result.tell(subtask_result)

    async def get_last_idle_time(self) -> Union[float, None]:
        return await self._task_manager_ref.get_last_idle_time()

    async def remove_tileables(self, tileable_keys: List[str]):
        return await self._task_manager_ref.remove_tileables(tileable_keys)

    async def get_map_reduce_info(
        self, task_id: str, map_reduce_id: int
    ) -> MapReduceInfo:
        return await self._task_manager_ref.get_map_reduce_info(task_id, map_reduce_id)

    async def save_task_info(self, task_info: Dict, path: str):
        await self._task_info_collector_ref.save_task_info.tell(task_info, path)

    async def collect_task_info_enabled(self):
        return await self._task_info_collector_ref.collect_task_info_enabled()



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import MARS_DATAFRAME_CALLABLES, MARS_DATAFRAME_MAGIC_METHODS


def _install():
    from ...core.adapter import (
        MARS_DATAFRAME_GROUPBY_TYPE,
        MARS_DATAFRAME_TYPE,
        MARS_INDEX_TYPE,
        MARS_SERIES_GROUPBY_TYPE,
        MARS_SERIES_TYPE,
        MarsDataFrame,
        MarsDataFrameGroupBy,
        MarsSeries,
        collect_cls_members,
        register_data_members,
        wrap_magic_method,
    )
    from ...core.data import DataRef, DataType
    from .core import (
        _register_from_mars_execution_conditions,
        _register_to_mars_execution_conditions,
        wrap_iteration_functions,
        wrap_user_defined_functions,
    )

    for method in MARS_DATAFRAME_MAGIC_METHODS:
        setattr(DataRef, method, wrap_magic_method(method))

    _register_to_mars_execution_conditions()
    _register_from_mars_execution_conditions()

    for cls in MARS_DATAFRAME_TYPE:
        register_data_members(
            DataType.dataframe, collect_cls_members(cls, DataType.dataframe)
        )
    for cls in MARS_SERIES_TYPE:
        register_data_members(
            DataType.series, collect_cls_members(cls, DataType.series)
        )
    for cls in MARS_INDEX_TYPE:
        register_data_members(DataType.index, collect_cls_members(cls, DataType.index))
    for cls in MARS_DATAFRAME_GROUPBY_TYPE:
        register_data_members(
            DataType.dataframe_groupby,
            collect_cls_members(cls, DataType.dataframe_groupby),
        )
    for cls in MARS_SERIES_GROUPBY_TYPE:
        register_data_members(
            DataType.series_groupby, collect_cls_members(cls, DataType.series_groupby)
        )

    # install DataFrame user defined functions:
    # DataFrame.apply
    # DataFrame.transform
    # DataFrame.map_chunk
    # DataFrame.cartesian_chunk
    dataframe_udfs = dict()
    dataframe_udfs["apply"] = wrap_user_defined_functions(
        MarsDataFrame.apply, "apply", DataType.dataframe
    )
    dataframe_udfs["transform"] = wrap_user_defined_functions(
        MarsDataFrame.apply, "transform", DataType.dataframe
    )
    dataframe_udfs["map_chunk"] = wrap_user_defined_functions(
        MarsDataFrame.map_chunk, "map_chunk", DataType.dataframe
    )
    dataframe_udfs["cartesian_chunk"] = wrap_user_defined_functions(
        MarsDataFrame.cartesian_chunk, "cartesian_chunk", DataType.dataframe
    )
    register_data_members(DataType.dataframe, dataframe_udfs)

    # install Serise user defined functions
    # Sereis.map
    register_data_members(
        DataType.series,
        dict(map=wrap_user_defined_functions(MarsSeries.map, "map", DataType.series)),
    )

    # install DataFrameGroupBy user defined functions:
    # GroupBy.apply
    # GroupBy.transform
    dataframe_groupby_udfs = dict()
    dataframe_groupby_udfs["apply"] = wrap_user_defined_functions(
        MarsDataFrameGroupBy.apply, "apply", DataType.dataframe_groupby
    )
    dataframe_groupby_udfs["transform"] = wrap_user_defined_functions(
        MarsDataFrameGroupBy.transform, "transform", DataType.dataframe_groupby
    )
    register_data_members(DataType.dataframe_groupby, dataframe_groupby_udfs)

    # install iteration functions
    # DataFrame.itertuples, DataFrame.iterrows
    dataframe_iteration_functions = dict()
    dataframe_iteration_functions["itertuples"] = wrap_iteration_functions(
        MarsDataFrame.itertuples, "itertuples", DataType.dataframe, True
    )
    dataframe_iteration_functions["iterrows"] = wrap_iteration_functions(
        MarsDataFrame.iterrows, "iterrows", DataType.dataframe, True
    )
    register_data_members(DataType.dataframe, dataframe_iteration_functions)

    # Series.items
    register_data_members(
        DataType.series,
        dict(
            items=wrap_iteration_functions(
                MarsSeries.items, "items", DataType.series, True
            )
        ),
    )



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from ..core.utils.fallback import unimplemented_func


def _install():
    """Nothing required for installing xgboost."""


def __dir__():
    from .mars_adapters import MARS_XGBOOST_CALLABLES

    return list(MARS_XGBOOST_CALLABLES.keys())


def __getattr__(name: str):
    from .mars_adapters import MARS_XGBOOST_CALLABLES

    if name in MARS_XGBOOST_CALLABLES:
        return MARS_XGBOOST_CALLABLES[name]
    else:
        try:
            import xgboost
        except ImportError:  # pragma: no cover
            xgboost = None

        if xgboost is not None:
            if not hasattr(xgboost, name):
                raise AttributeError(name)
            else:  # pragma: no cover
                if inspect.ismethod(getattr(xgboost, name)):
                    return unimplemented_func
                else:
                    raise AttributeError(name)
